"""
This file contains the basic dataset functions.
"""
from typing import Dict, List, Optional, Union

import pandas as pd

from hyfi.utils.logging import LOGGING
from hyfi.utils.types import DictLike, ListLike

logger = LOGGING.getLogger(__name__)


class DSBasic:
    @staticmethod
    def dataframe_select_columns(
        data: pd.DataFrame,
        columns: Union[List[str], str],
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Select columns from a dataframe.

        Args:
            data (pd.DataFrame): The dataframe to select columns from.
            columns (List[str]): The columns to select.
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.DataFrame: The dataframe with the columns selected.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
            >>> DSBasic.dataframe_select_columns(data, columns=["a"])
               a
            0  1
            1  2
            2  3
        """
        columns = [columns] if isinstance(columns, str) else columns
        columns = [c for c in columns if c in data.columns]
        if verbose:
            logger.info("Selecting columns: %s", columns)
        return data[columns]

    @staticmethod
    def dataframe_drop_columns(
        data: pd.DataFrame,
        columns: Union[List[str], str],
        level: Optional[int] = None,
        errors: str = "raise",
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Drop columns from a dataframe.

        Args:
            data (pd.DataFrame): The dataframe to drop columns from.
            columns (List[str]): The columns to drop.
            level (int, optional): The level to drop columns from. Defaults to None.
            errors (str, optional): Whether to raise an error if the column does not exist. Defaults to "raise".
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.DataFrame: The dataframe with the columns dropped.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
            >>> DSBasic.dataframe_drop_columns(data, columns=["a"])
                b
            0   4
            1   5
            2   6
        """
        if verbose:
            logger.info("Dropping columns %s", columns)
        if isinstance(columns, str):
            columns = [columns]
        data.drop(columns=columns, level=level, errors=errors, inplace=True)
        return data

    @staticmethod
    def dataframe_drop(
        data: pd.DataFrame,
        labels: Union[List[str], str],
        axis: int = 1,
        level: Optional[int] = None,
        errors: str = "raise",
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Drop columns from a dataframe.

        Args:
            data (pd.DataFrame): The dataframe to drop columns from.
            labels (List[str]): The columns or index to drop.
            axis (int, optional): The axis to drop from. Defaults to 1.
            level (int, optional): The level to drop columns from. Defaults to None.
            errors (str, optional): Whether to raise an error if the column does not exist. Defaults to "raise".
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.DataFrame: The dataframe with the columns dropped.
        """
        if verbose:
            logger.info("Dropping labels %s", labels)
        if isinstance(labels, str):
            labels = [labels]
        data.drop(labels=labels, axis=axis, level=level, errors=errors, inplace=True)
        return data

    @staticmethod
    def dataframe_split_str_column(
        data: pd.DataFrame,
        column: str,
        sep: str,
        new_column_name: Optional[str] = None,
        drop_old_column: bool = False,
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Split a column of strings into a new column.

        Args:
            data (pd.DataFrame): The dataframe to split.
            column (str): The column to split.
            sep (str): The string to split on.
            new_column_name (str): The name of the new column.
            drop_old_column (bool, optional): Whether to drop the old column. Defaults to False.
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.Dataframe: The dataframe with the new column.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": ["1,2", "3,4", "5,6"]})
            >>> DSBasic.dataframe_split_str_column(data, column="a", sep=",")
                a_1
            0   1
            1   3
            2   5
        """
        if new_column_name is None:
            new_column_name = column
            drop_old_column = False
        if verbose:
            logger.info(
                "Splitting column %s on %s into new column %s",
                column,
                sep,
                new_column_name,
            )
        data[new_column_name] = data[column].str.split(sep)
        if drop_old_column:
            data.drop(column, axis=1, inplace=True)
        return data

    @staticmethod
    def dataframe_combine_str_columns(
        data: pd.DataFrame,
        columns: List[str],
        sep: str,
        fillna: Optional[str] = None,
        new_column_name: Optional[str] = None,
        drop_old_columns: bool = False,
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Combine columns of strings into a new column.

        Args:
            data (pd.DataFrame): The dataframe to combine.
            columns (List[str]): The columns to combine.
            sep (str): The string to combine on.
            fillna (str, optional): The string to fill NaN values with. Defaults to None.
            new_column_name (str): The name of the new column.
            drop_old_columns (bool, optional): Whether to drop the old columns. Defaults to False.
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.Dataframe: The dataframe with the new column.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": ["1", "3", "5"], "b": ["2", "4", "6"]})
            >>> DSBasic.dataframe_combine_str_columns(data, columns=["a", "b"], sep=",")
                a_b
            0   1,2
            1   3,4
            2   5,6

        """
        if new_column_name is None:
            new_column_name = "_".join(columns)
            drop_old_columns = False
        if verbose:
            logger.info(
                "Combining columns %s into new column %s",
                columns,
                new_column_name,
            )
        if fillna:
            data[new_column_name] = data[columns].fillna(fillna).agg(sep.join, axis=1)
        else:
            data[new_column_name] = data[columns].agg(sep.join, axis=1)
        if drop_old_columns:
            data.drop(columns, axis=1, inplace=True)
        return data

    @staticmethod
    def dataframe_eval_columns(
        data: pd.DataFrame,
        expressions: Union[Dict[str, str], List[str]],
        engine: str = "python",
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Evaluate columns in a dataframe.

        Args:
            data (pd.DataFrame): The dataframe to evaluate.
            expressions (Union[Dict[str, str], List[str]]): The expressions to evaluate.
            engine (str, optional): The engine to use. Defaults to "python".
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.DataFrame: The dataframe with the evaluated columns.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
            >>> DSBasic.dataframe_eval_columns(data, expressions={"c": "a + b"})
                a   b   c
            0   1   4   5
            1   2   5   7
            2   3   6   9

        """
        if isinstance(expressions, DictLike):
            for column in expressions:
                if verbose:
                    logger.info("Evaluating column %s", column)
                data[column] = data.eval(expressions[column], engine=engine)
        elif isinstance(expressions, ListLike):
            for expression in expressions:
                if verbose:
                    logger.info("Evaluating expression %s", expression)
                data = data.eval(expression, engine=engine)
        return data

    @staticmethod
    def dataframe_eval_columns_with_pd_eval(
        data: pd.DataFrame,
        expressions: Union[Dict[str, str], List[str]],
        engine: str = "python",
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Evaluate columns in a dataframe using the pandas eval function.

        Args:
            data (pd.DataFrame): The dataframe to evaluate.
            expressions (Union[Dict[str, str], List[str]]): The expressions to evaluate.
            engine (str, optional): The engine to use. Defaults to "python".
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.DataFrame: The dataframe with the evaluated columns.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
            F>>> DSBasic.dataframe_eval_columns_with_pd_eval(data, expressions={"c": "data.a + data.b"})
                a   b   c
            0   1   4   5
            1   2   5   7
            2   3   6   9

        """
        if isinstance(expressions, DictLike):
            for column in expressions:
                if verbose:
                    logger.info("Evaluating column %s", column)
                data[column] = pd.eval(expressions[column], engine=engine, target=data)
        elif isinstance(expressions, ListLike):
            for expression in expressions:
                if verbose:
                    logger.info("Evaluating expression %s", expression)
                data = pd.eval(expression, engine=engine, target=data)
        return data

    # @staticmethod
    # def dataframe_eval_columns_with_eval(
    #     data: pd.DataFrame,
    #     expressions: Dict[str, str],
    #     verbose: bool = False,
    # ) -> pd.DataFrame:
    #     """
    #     Evaluate columns in a dataframe using the built-in eval function.

    #     Args:
    #         data (pd.DataFrame): The dataframe to evaluate.
    #         expressions (Union[Dict[str, str], List[str]]): The expressions to evaluate.
    #         engine (str, optional): The engine to use. Defaults to "python".
    #         verbose (bool, optional): Whether to print verbose output. Defaults to False.

    #     Returns:
    #         pd.DataFrame: The dataframe with the evaluated columns.

    #                 Examples:
    #         >>> import pandas as pd
    #         >>> from hyfi.utils.datasets.basic import DSBasic
    #         >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    #         F>>> DSBasic.dataframe_eval_columns_with_eval(data, expressions={"c": "data.a + data.b"})
    #             a   b   c
    #         0   1   4   5
    #         1   2   5   7
    #         2   3   6   9
    #     """
    #     for column in expressions:
    #         if verbose:
    #             logger.info("Evaluating column %s", column)
    #         data[column] = eval(expressions[column])
    #     return data

    @staticmethod
    def dataframe_print_head_and_tail(
        data: pd.DataFrame,
        num_heads: Optional[int] = 5,
        num_tails: Optional[int] = 5,
        columns: Optional[Union[List[str], str]] = None,
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Print the head and tail of a dataframe.

        Args:
            data (pd.DataFrame): The dataframe to print.
            num_heads (int, optional): The number of rows to print from the head. Defaults to 5.
            num_tails (int, optional): The number of rows to print from the tail. Defaults to 5.
            verbose (bool, optional): Whether to print verbose output. Defaults to False.

        Returns:
            pd.DataFrame: The dataframe.

        Examples:
            >>> import pandas as pd
            >>> from hyfi.utils.datasets.basic import DSBasic
            >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
            >>> DSBasic.dataframe_print_head_and_tail(data, num_heads=1, num_tails=1)
                a   b
            0   1   4
            2   3   6
        """
        if not columns:
            columns = data.columns
        elif isinstance(columns, str):
            columns = [columns]
        if verbose:
            logger.info("Printing head and tail of dataframe")
        if num_heads:
            if verbose:
                logger.info("Head:")
            print(data[columns].head(num_heads))
        if num_tails:
            if verbose:
                logger.info("Tail:")
            print(data[columns].tail(num_tails))
        return data



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import inspect
import sys
from typing import Any, Tuple

from typing_extensions import Final

UNKNOWN_NAME: Final[str] = "<unknown>"
COMMON_MODULES_WITH_OBFUSCATED_IMPORTS: Tuple[str, ...] = (
    "random",
    "numpy",
    "numpy.random",
    "jax.numpy",
    "jax",
    "torch",
)

# signature param-types
_POSITIONAL_ONLY: Final = inspect.Parameter.POSITIONAL_ONLY
_POSITIONAL_OR_KEYWORD: Final = inspect.Parameter.POSITIONAL_OR_KEYWORD
_VAR_POSITIONAL: Final = inspect.Parameter.VAR_POSITIONAL
_KEYWORD_ONLY: Final = inspect.Parameter.KEYWORD_ONLY
_VAR_KEYWORD: Final = inspect.Parameter.VAR_KEYWORD


def safe_name(obj: Any, repr_allowed: bool = True) -> str:
    """Tries to get a descriptive name for an object. Returns '<unknown>`
    instead of raising - useful for writing descriptive/dafe error messages."""

    if hasattr(obj, "__name__"):
        return obj.__name__

    return repr(obj) if repr_allowed and hasattr(obj, "__repr__") else UNKNOWN_NAME


def is_classmethod(obj: Any) -> bool:
    """
    Check if an object is a classmethod.

    https://stackoverflow.com/a/19228282/6592114

    Credit to: Martijn Pieters
    License: CC BY-SA 4.0 (free to copy/redistribute/remix/transform)
    """

    if not inspect.ismethod(obj):
        return False

    bound_to = getattr(obj, "__self__", None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = safe_name(obj)

    if name == UNKNOWN_NAME:  # pragma: no cover
        return False

    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False  # pragma: no cover


def get_obj_path(obj: Any) -> str:
    name = safe_name(obj, repr_allowed=False)

    if name == UNKNOWN_NAME:
        raise AttributeError(f"{obj} does not have a `__name__` attribute")

    module = getattr(obj, "__module__", None)
    qualname = getattr(obj, "__qualname__", None)

    if (qualname is not None and "<" in qualname) or module is None:
        # NumPy's ufuncs do not have an inspectable `__module__` attribute, so we
        # check to see if the object lives in NumPy's top-level namespace.
        #
        # or..
        #
        # Qualname produced a name from a local namespace.
        # E.g. jax.numpy.add.__qualname__ is '_maybe_bool_binop.<locals>.fn'
        # Thus we defer to the name of the object and look for it in the
        # top-level namespace of the known suspects
        #
        # or...
        #
        # module is None, which is apparently a thing..:
        # __module__ is None for both numpy.random.rand and random.random
        #

        # don't use qualname for obfuscated paths
        for new_module in COMMON_MODULES_WITH_OBFUSCATED_IMPORTS:
            if getattr(sys.modules.get(new_module), name, None) is obj:
                module = new_module
                break
        else:
            raise ModuleNotFoundError(f"{name} is not importable")

    if not is_classmethod(obj):
        return f"{module}.{name}"
    else:
        # __qualname__ reflects name of class that originally defines classmethod.
        # Does not point to child in case of inheritance.
        #
        # obj.__self__ -> parent object
        # obj.__name__ -> name of classmethod
        return f"{get_obj_path(obj.__self__)}.{obj.__name__}"


def get_sig_obj(target):
    if not inspect.isclass(target):
        return target

    # This implements the same method prioritization as
    # `inspect.signature` for Python >= 3.9.1
    if "__new__" in target.__dict__:
        return target.__new__
    if "__init__" in target.__dict__:
        return target.__init__

    if len(target.__mro__) > 2:
        for parent in target.__mro__[1:-1]:
            if "__new__" in parent.__dict__:
                return target.__new__
            elif "__init__" in parent.__dict__:
                return target.__init__
    return getattr(target, "__init__", target)



import json
import os
import time
from dataclasses import asdict, dataclass
from typing import Optional, Set

from .common import PathOrStr


@dataclass
class Meta:
    """
    Any resource that is downloaded to - or extracted in - the cache directory will
    have a meta JSON file written next to it, which corresponds to an instance
    of this class.

    In older versions of AllenNLP, this meta document just had two fields: 'url' and
    'etag'. The 'url' field is now the more general 'resource' field, but these old
    meta files are still compatible when a `Meta` is instantiated with the `.from_path()`
    class method.
    """

    resource: str
    """
    URL or normalized path to the resource.
    """

    cached_path: str
    """
    Path to the corresponding cached version of the resource.
    """

    creation_time: float
    """
    The unix timestamp of when the corresponding resource was cached or extracted.
    """

    size: int = 0
    """
    The size of the corresponding resource, in bytes.
    """

    etag: Optional[str] = None
    """
    Optional ETag associated with the current cached version of the resource.
    """

    extraction_dir: bool = False
    """
    Does this meta corresponded to an extraction directory?
    """

    @classmethod
    def new(
        cls,
        resource: PathOrStr,
        cached_path: PathOrStr,
        *,
        etag: Optional[str] = None,
        extraction_dir: bool = False,
    ) -> "Meta":
        return cls(
            resource=str(resource),
            cached_path=str(cached_path),
            creation_time=time.time(),
            size=cls.get_resource_size(cached_path),
            etag=etag,
            extraction_dir=extraction_dir,
        )

    def to_file(self) -> None:
        with open(f"{self.cached_path}.json", "w") as meta_file:
            json.dump(asdict(self), meta_file)

    @classmethod
    def from_path(cls, path: PathOrStr) -> "Meta":
        path = str(path)
        with open(path) as meta_file:
            data = json.load(meta_file)
            # For backwards compat:
            if "resource" not in data:
                data["resource"] = data.pop("url")
            if "creation_time" not in data:
                data["creation_time"] = os.path.getmtime(path[:-5])
            if "extraction_dir" not in data and path.endswith("-extracted.json"):
                data["extraction_dir"] = True
            if "cached_path" not in data:
                data["cached_path"] = path[:-5]
            if "size" not in data:
                data["size"] = cls.get_resource_size(data["cached_path"])
        return cls(**data)

    @staticmethod
    def get_resource_size(path: PathOrStr) -> int:
        """
        Get the size of a file or directory.
        """
        if os.path.isfile(path):
            return os.path.getsize(path)
        inodes: Set[int] = set()
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(str(path)):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                # skip if it is symbolic link or the same as a file we've already accounted
                # for (this could happen with hard links).
                inode = os.stat(fp).st_ino
                if not os.path.islink(fp) and inode not in inodes:
                    inodes.add(inode)
                    total_size += os.path.getsize(fp)
        return total_size



"""
Save datasets to disk
"""
import os
from os import PathLike
from pathlib import Path
from typing import Optional, Sequence, Union

import pandas as pd
from datasets.dataset_dict import DatasetDict, IterableDatasetDict

from hyfi.utils.contexts import elapsed_timer
from hyfi.utils.logging import LOGGING

from .types import DatasetLikeType
from .utils import DSUtils

logger = LOGGING.getLogger(__name__)


class DSSave:
    def __init__(self):
        pass

    @staticmethod
    def save_dataframes(
        data: Union[pd.DataFrame, dict],
        data_file: str,
        data_dir: Optional[str] = None,
        columns: Optional[Sequence[str]] = None,
        index: bool = False,
        filetype: Optional[str] = "parquet",
        suffix: Optional[str] = None,
        verbose: bool = False,
        **kwargs,
    ):
        """Save data to a file"""
        if data_file is None:
            raise ValueError("filename must be specified")
        fileinfo = os.path.splitext(data_file)
        data_file = fileinfo[0]
        filetype = fileinfo[1] if len(fileinfo) > 1 else filetype
        filetype = "." + filetype.replace(".", "")
        if suffix:
            data_file = f"{data_file}-{suffix}{filetype}"
        else:
            data_file = f"{data_file}{filetype}"
        filepath = os.path.join(data_dir, data_file) if data_dir else data_file
        data_dir = os.path.dirname(filepath)
        data_file = os.path.basename(filepath)
        os.makedirs(data_dir, exist_ok=True)

        if isinstance(data, dict):
            for k, v in data.items():
                DSSave.save_dataframes(
                    v,
                    data_file,
                    data_dir=data_dir,
                    columns=columns,
                    index=index,
                    filetype=filetype,
                    suffix=k,
                    verbose=verbose,
                    **kwargs,
                )
        elif DSUtils.is_dataframe(data):
            logger.info("Saving dataframe to %s", Path(filepath).absolute())
            if isinstance(columns, list):
                columns = [c for c in columns if c in data.columns]
                data = data[columns]
            with elapsed_timer(format_time=True) as elapsed:
                if "csv" in filetype or "tsv" in filetype:
                    data.to_csv(filepath, index=index)
                elif "parquet" in filetype:
                    compression = kwargs.get("compression", "gzip")
                    engine = kwargs.get("engine", "pyarrow")
                    if engine not in ["pyarrow", "fastparquet"]:
                        engine = "auto"
                    data.to_parquet(filepath, compression=compression, engine=engine)  # type: ignore
                else:
                    raise ValueError("filetype must be .csv or .parquet")
                if verbose:
                    logger.info(" >> elapsed time to save data: %s", elapsed())
        else:
            raise ValueError(f"Unsupported data type: {type(data)}")

    @staticmethod
    def save_dataset_to_disk(
        dset: DatasetLikeType,
        dataset_path: PathLike,
        max_shard_size: Optional[Union[str, int]] = None,
        num_shards: Optional[int] = None,
        num_proc: Optional[int] = None,
        storage_options: Optional[dict] = None,
        verbose: bool = False,
        **kwargs,
    ) -> DatasetLikeType:
        """Save a dataset or a dataset dict to the filesystem."""
        dset.save_to_disk(
            dataset_path,
            max_shard_size=max_shard_size,
            num_shards=num_shards,
            num_proc=num_proc,
            storage_options=storage_options,
        )
        logger.info("Dataset saved to %s.", dataset_path)
        if verbose:
            if isinstance(dset, DatasetDict or IterableDatasetDict):
                for split in dset:
                    logger.info("Split: %s", split)
                    logger.info("Dataset features: %s", dset[split].features)
                    logger.info("Number of records: %s", len(dset[split]))
            else:
                logger.info("Dataset features: %s", dset.features)
                logger.info("Number of records: %s", len(dset))

        return dset



"""Image utils."""
import io
import os
import platform
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import font_manager, rc
from PIL import Image, ImageDraw, ImageFont

from hyfi.utils.iolibs import IOLIBs
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class GUTILs:
    """Image utils."""

    @staticmethod
    def get_grid_size(
        num_images: int,
        num_cols: Optional[int] = None,
        num_rows: Optional[int] = None,
        truncate: bool = True,
        fit_to_grid: bool = True,
    ) -> Tuple[int, int, int]:
        """Get number of rows and columns for a grid of images.

        Args:
            num_images (int): Number of images.
            num_cols (int, optional): Number of columns. Defaults to None.
            num_rows (int, optional): Number of rows. Defaults to None.
            truncate (bool, optional): Truncate number of images to fit grid. Defaults to True. If False, the number of images will be increased to fit the grid.

        Returns:
            Tuple[int, int, int]: Number of images, columns and rows.
        """
        if not num_cols and num_images % num_cols != 0 and truncate:
            num_images = (num_images // num_cols) * num_cols
        if not num_cols or num_cols > num_images or num_cols < 1:
            num_cols = num_images // 2
        num_rows = num_images // num_cols

        return num_images, num_cols, num_rows

    @staticmethod
    def scale_image(
        image: Image.Image,
        max_width: Optional[int] = None,
        max_height: Optional[int] = None,
        max_pixels: Optional[int] = None,
        scale: float = 1.0,
        resize_to_multiple_of: Optional[int] = None,
        resample: int = Image.LANCZOS,
    ) -> Image.Image:
        """Scale image to have at most `max_pixels` pixels.

        Args:
            image: PIL image.
            max_width: Maximum width.
            max_height: Maximum height.
            max_pixels: Maximum number of pixels.
            scale: Scale factor.
            resize_to_multiple_of: Resize to multiple of this value.
            resample: Resampling filter.

        Returns:
            PIL image.
        """

        w, h = image.size

        if not max_width and max_height:
            max_width = int(w * max_height / h)
        elif not max_height and max_width:
            max_height = int(h * max_width / w)
        else:
            max_width = max_width or w
            max_height = max_height or h

        if max_width and max_height:
            max_pixels = max_width * max_height

        scale = np.sqrt(max_pixels / (w * h)) if max_pixels > 0 else scale or 1.0
        max_width = int(w * scale)
        max_height = int(h * scale)
        if resize_to_multiple_of:
            max_width = (max_width // resize_to_multiple_of) * resize_to_multiple_of
            max_height = (max_height // resize_to_multiple_of) * resize_to_multiple_of

        if scale < 1.0 or w > max_width or h > max_height:
            image = image.resize((max_width, max_height), resample=resample)  # type: ignore
        return image

    @staticmethod
    def load_image_as_ndarray(
        image_or_uri: Union[str, Path, Image.Image],
    ) -> np.ndarray:
        """Load image from file or URI."""
        return np.asarray(GUTILs.load_image(image_or_uri))

    @staticmethod
    def load_image(
        image_or_uri: Union[str, Path, Image.Image],
        max_width: Optional[int] = None,
        max_height: Optional[int] = None,
        max_pixels: Optional[int] = None,
        scale: float = 1.0,
        resize_to_multiple_of: Optional[int] = None,
        crop_box: Optional[Tuple[int, int, int, int]] = None,
        mode: str = "RGB",
        **read_kwargs,
    ) -> Image.Image:
        """Load image from file or URI."""
        from PIL import Image

        if isinstance(image_or_uri, Image.Image):
            img = image_or_uri.convert(mode)
        elif Path(image_or_uri).is_file():
            img = Image.open(image_or_uri).convert(mode)
        else:
            img = Image.open(
                io.BytesIO(IOLIBs.read(image_or_uri, **read_kwargs))
            ).convert(mode)
        img = GUTILs.scale_image(
            img,
            max_width=max_width,
            max_height=max_height,
            max_pixels=max_pixels,
            scale=scale,
            resize_to_multiple_of=resize_to_multiple_of,
        )
        if crop_box is not None:
            img = img.crop(crop_box)
        return img

    @staticmethod
    def load_images(
        images_or_uris: List[Union[str, Image.Image]],
        max_width: Optional[int] = None,
        max_height: Optional[int] = None,
        max_pixels: Optional[int] = None,
        scale: float = 1.0,
        resize_to_multiple_of: Optional[int] = None,
        crop_to_min_size: bool = False,
        mode: str = "RGB",
        **kwargs,
    ) -> List[Image.Image]:
        """Load images from files or URIs."""
        imgs = [
            GUTILs.load_image(
                image_or_uri,
                max_width=max_width,
                max_height=max_height,
                max_pixels=max_pixels,
                scale=scale,
                resize_to_multiple_of=resize_to_multiple_of,
                mode=mode,
                **kwargs,
            )
            for image_or_uri in images_or_uris
        ]
        if crop_to_min_size:
            min_width = min(img.width for img in imgs)
            min_height = min(img.height for img in imgs)
            if resize_to_multiple_of is not None:
                min_width = (min_width // resize_to_multiple_of) * resize_to_multiple_of
                min_height = (
                    min_height // resize_to_multiple_of
                ) * resize_to_multiple_of
            imgs = [img.crop((0, 0, min_width, min_height)) for img in imgs]

        return imgs

    @staticmethod
    def get_image_font(
        fontname: Optional[str] = None,
        fontsize: int = 12,
    ) -> Optional[ImageFont.ImageFont]:
        """Get font for PIL image."""
        fontname, fontpath = GUTILs.get_plot_font(
            fontname=fontname,
            set_font_for_matplot=False,
        )
        return ImageFont.truetype(fontpath, fontsize) if fontpath else None

    @staticmethod
    def get_default_system_font(
        fontname: Optional[str] = None,
        fontpath: Optional[str] = None,
        lang: str = "ko",
        verbose: bool = False,
    ) -> Tuple[str, str]:
        if platform.system() == "Darwin":
            default_fontname = "AppleGothic.ttf" if lang == "ko" else "Arial.ttf"
            fontname = fontname or default_fontname
            fontpath = os.path.join("/System/Library/Fonts/Supplemental/", fontname)
        elif platform.system() == "Windows":
            default_fontname = "malgun.ttf" if lang == "ko" else "arial.ttf"
            fontname = fontname or default_fontname
            fontpath = os.path.join("c:/Windows/Fonts/", fontname)
        elif platform.system() == "Linux":
            default_fontname = "NanumGothic.ttf" if lang == "ko" else "DejaVuSans.ttf"
            fontname = fontname or default_fontname
            if fontname.lower().startswith("nanum"):
                fontpath = os.path.join("/usr/share/fonts/truetype/nanum/", fontname)
            else:
                fontpath = os.path.join("/usr/share/fonts/truetype/", fontname)
        if fontpath and not Path(fontpath).is_file():
            paths = GUTILs.find_font_file(fontname)
            fontpath = paths[0] if len(paths) > 0 else ""
        if verbose:
            logger.info(f"Font path: {fontpath}")
        return fontname, fontpath

    @staticmethod
    def get_plot_font(
        fontpath: Optional[str] = None,
        fontname: Optional[str] = None,
        lang: str = "en",
        set_font_for_matplot: bool = True,
        verbose: bool = False,
    ) -> Tuple[str, str]:
        """Get font for plot

        Args:
            fontpath: Font file path
            fontname: Font name
            lang: Language
            set_font_for_matplot: Set font for matplot
            verbose: Verbose mode

        Returns:
            Tuple of font name and font path

        """
        if fontname and not fontname.endswith(".ttf"):
            fontname += ".ttf"
        if not fontpath:
            fontname, fontpath = GUTILs.get_default_system_font(
                fontname, fontpath, lang, verbose
            )

        if fontpath and Path(fontpath).is_file():
            font_manager.fontManager.addfont(fontpath)
            fontname = font_manager.FontProperties(fname=fontpath).get_name()  # type: ignore

            if set_font_for_matplot and fontname:
                rc("font", family=fontname)
                plt.rcParams["axes.unicode_minus"] = False
                if verbose:
                    font_family = plt.rcParams["font.family"]
                    logger.info(f"font family: {font_family}")
            if verbose:
                logger.info(f"font name: {fontname}")
        else:
            logger.warning(f"Font file does not exist at {fontpath}")
            fontname = ""
            fontpath = ""
            if platform.system() == "Linux":
                font_install_help = """
                apt install fontconfig
                apt install fonts-nanum
                fc-list | grep -i nanum
                """
                print(font_install_help)
        return fontname, fontpath

    @staticmethod
    def find_font_file(query: str) -> List[str]:
        """Find font file by query string"""
        return list(
            filter(
                lambda path: query in os.path.basename(path),
                font_manager.findSystemFonts(),
            )
        )

    @staticmethod
    def convert_figure_to_image(
        fig: plt.Figure,
        dpi: int = 300,
    ):
        """Convert a Matplotlib figure to a PIL Image and return it"""
        buf = io.BytesIO()
        fig.savefig(buf, format="png", dpi=dpi, bbox_inches="tight", pad_inches=0)
        buf.seek(0)
        return Image.open(buf)

    @staticmethod
    def print_filename_on_image(
        image_or_uri,
        print_filename: bool = False,
        filename: Optional[str] = None,
        filename_offset: Tuple[float, float] = (5, 5),
        fontname: Optional[str] = None,
        fontsize: int = 12,
        fontcolor: Optional[str] = None,
    ) -> Image.Image:
        """
        Convert an image to a PIL Image.

        Args:
            image_or_uri: The image or image URI.
            print_filename: Whether to print the filename on the image.
            filename: The filename to show on the image.
            filename_offset: The offset of the filename from the top left corner.
            fontname: The font name to use for the filename.
            fontsize: The font size to use for the filename.
            fontcolor: The font color to use for the filename.

        Returns:
            The PIL Image.
        """
        img = GUTILs.load_image(image_or_uri)
        if isinstance(image_or_uri, str) and filename is None:
            filename = os.path.basename(image_or_uri)
        if print_filename and filename:
            font = GUTILs.get_image_font(fontname, fontsize)
            draw = ImageDraw.Draw(img)
            draw.text(filename_offset, filename, font=font, fill=fontcolor)

        # img = img.convert("RGB")
        # img = np.array(img)
        return img

    @staticmethod
    def get_ticks_from_lim(
        lim: Tuple[float, float],
        num_ticks: int,
    ) -> np.ndarray:
        """
        Get n evenly spaced ticks from lim.

        Args:
            lim: Tuple of (min, max) values.
            num_ticks: Number of ticks to get.

        Returns:
            Array of ticks.
        """
        ticks = np.linspace(lim[0], lim[1], num_ticks + 1)
        ticks = ticks - (ticks[1] - ticks[0]) / 2
        ticks[0] = lim[0]
        return ticks

    @staticmethod
    def save_adjusted_subplots(
        fig: plt.Figure,
        output_file: str,
        left: float = 0.1,
        right: float = 0.9,
        bottom: float = 0.1,
        top: float = 0.9,
        wspace: float = 0,
        hspace: float = 0,
        tight_layout: bool = True,
        bbox_inches: str = "tight",
        pad_inches: float = 0,
        transparent: bool = True,
        dpi: int = 300,
        verbose: bool = True,
    ):
        """Save subplots after adjusting the figure

        Args:
            fig: Figure
            output_file: Output file path
            left: Left
            right: Right
            bottom: Bottom
            top: Top
            wspace: Wspace
            hspace: Hspace
            tight_layout: Tight layout
            bbox_inches: Bbox inches
            pad_inches: Pad inches
            transparent: Transparent
            dpi: DPI
            verbose: Verbose mode
        """

        # make the figure look better
        fig.subplots_adjust(
            left=left,
            right=right,
            bottom=bottom,
            top=top,
            wspace=wspace,
            hspace=hspace,
        )
        if tight_layout:
            fig.tight_layout()

        Path(output_file).parent.mkdir(parents=True, exist_ok=True)
        fig.savefig(
            output_file,
            dpi=dpi,
            bbox_inches=bbox_inches,
            pad_inches=pad_inches,
            transparent=transparent,
        )
        if verbose:
            logger.info("Saved subplots to %s", output_file)



from pathlib import Path

from hyfi.utils.logging import LOGGING

from .base import BasePathConfig

logger = LOGGING.getLogger(__name__)


class ProjectPathConfig(BasePathConfig):
    _config_name_: str = "__project__"

    # internal paths for hyfi
    home: str
    hyfi: str
    resources: str
    runtime: str
    # global paths
    global_hyfi_root: str = "."
    global_workspace_name: str = ".hyfi"
    # project specific paths
    project_name: str
    project_root: str = "."
    project_workspace_name: str = "workspace"

    @property
    def home_dir(self) -> Path:
        """
        Return the path to the home directory.

        Returns:
            path to the home directory
        """
        return Path(self.home).absolute() if self.home else Path.home()

    @property
    def global_root_dir(self) -> Path:
        """
        Create and return the path to the hyfi directory.

        Returns:
            path to the hyfi directory
        """
        path_ = Path(self.global_hyfi_root)
        if not path_.is_absolute():
            path_ = self.home_dir / path_
        return path_

    @property
    def global_workspace_dir(self) -> Path:
        """
        Create and return the path to the glboal workspace directory.

        Returns:
            path to the global workspace directory
        """
        return self.global_root_dir / self.global_workspace_name

    @property
    def global_archive_dir(self) -> Path:
        """
        Create and return the path to the global archive directory.

        Returns:
            path to the global archive directory
        """
        return self.get_path("archives", self.global_workspace_dir)

    @property
    def glboal_dataset_dir(self) -> Path:
        """
        Create and return the path to the global datasets directory.

        Returns:
            path to the global datasets directory
        """
        return self.get_path("datasets", self.global_workspace_dir)

    @property
    def global_model_dir(self) -> Path:
        """
        Create and return the path to the global models directory.

        Returns:
            path to the global models directory
        """
        return self.get_path("models", self.global_workspace_dir)

    @property
    def glboal_module_dir(self) -> Path:
        """
        Create and return the path to the global modules directory.

        Returns:
            path to the global modules directory
        """
        return self.get_path("modules", self.global_workspace_dir)

    @property
    def global_library_dir(self) -> Path:
        """
        Create and return the path to the global library directory.

        Returns:
            path to the global library directory
        """
        return self.get_path("library", self.global_workspace_dir)

    @property
    def glboal_log_dir(self) -> Path:
        """
        Create and return the path to the global log directory.

        Returns:
            path to the global log directory
        """
        return self.get_path("logs", self.global_workspace_dir)

    @property
    def global_cache_dir(self) -> Path:
        """
        Create and return the path to the global cache directory.

        Returns:
            path to the global cache directory
        """
        return self.get_path("cache", self.global_workspace_dir)

    @property
    def global_tmp_dir(self) -> Path:
        """
        Create and return the path to the global tmp directory.

        Returns:
            path to the global tmp directory
        """
        return self.get_path("tmp", self.global_workspace_dir)

    @property
    def name(self) -> str:
        """
        Returns the name of the path configuration.
        """
        return self.project_name

    @property
    def root_dir(self) -> Path:
        """
        Create and return the path to the project directory.

        Returns:
            path to the project directory
        """
        path_ = Path(self.project_root)
        return path_.absolute()

    @property
    def workspace_dir(self) -> Path:
        """
        Create and return the path to the project workspace directory.

        Returns:
            path to the project workspace directory
        """
        return self.root_dir / self.project_workspace_name



"""
    HiFY Base Model Class
"""
from pathlib import Path
from typing import Any, Dict, Optional, Set

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, model_validator

from hyfi.core import global_hyfi
from hyfi.utils.logging import LOGGING

from .composer import Composer
from .generator import sanitized_default_value

logger = LOGGING.getLogger(__name__)


class BaseModel(PydanticBaseModel):
    """
    Base class for all Pydantic models.

    Attributes:
        _config_name_ (str): The name of the model.
        _config_group_ (str): The group of the model.
        _auto_populate_ (bool): Whether to auto-populate the model with defaults from the config.
        _auto_generate_ (bool): Whether to auto-generate the config for the model.

    """

    _config_name_: str = "__init__"
    _config_group_: str = "/composer"
    _auto_populate_: bool = False
    _auto_generate_: bool = False
    _exclude_: Set[str] = set()
    _exclude_keys_ = {
        "_exclude_keys_": True,
        "_target_": True,
        "_config_name_": True,
        "_config_group_": True,
        "_auto_populate_": True,
        "_auto_generate_": True,
        "_exclude_": True,
    }

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        validate_assignment=False,
    )  # type: ignore

    @model_validator(mode="before")
    def validate_model_config_before(cls, data):
        # logger.debug("Validating model config before validating each field.")
        _auto_populate_ = data.get("_auto_populate_", getattr(cls._auto_populate_, "default", False))  # type: ignore
        if not _auto_populate_:
            if global_hyfi.verbosity > 1:
                logger.debug("Auto-populate is disabled for class `%s`.", cls.__name__)
            return data
        _config_name_ = data.get("_config_name_", getattr(cls._config_name_, "default", "__init__"))  # type: ignore
        _config_group_ = data.get("_config_group_", getattr(cls._config_group_, "default"))  # type: ignore
        _class_name_ = cls.__name__  # type: ignore
        if not _config_group_:
            if global_hyfi.verbosity > 0:
                logger.debug("There is no config group specified.")
            return data
        # Initialize the config with the given config_name.
        if global_hyfi.verbosity > 1:
            logger.info(
                "Composing `%s` class with `%s` config in `%s` group.",
                _class_name_,
                _config_name_,
                _config_group_,
            )
        config_group = f"{_config_group_}={_config_name_}"
        cfg = Composer.compose_as_dict(
            config_group=config_group,
            config_data=data,
            throw_on_compose_failure=False,
        )
        data = Composer.update(cfg, data)
        # Exclude any attributes specified in the class's `exclude` list.
        exclude = getattr(cls._exclude_, "default", set())  # type: ignore
        for name in exclude:
            if name in data:
                logger.debug("Excluding `%s` from the config.", name)
                del data[name]  # type: ignore
        return data

    @property
    def kwargs(self) -> Dict[str, Any]:
        """
        Returns the model as a dictionary excluding any keys specified in the class's `exclude_keys` list.
        """
        return self.model_dump(exclude=self._exclude_keys_)

    @property
    def config_name(self) -> str:
        """
        Returns the name of the model.
        """
        return self._config_name_

    @property
    def config_group(self) -> str:
        """
        Returns the group of the model.
        """
        return self._config_group_

    @property
    def auto_populate(self) -> bool:
        """
        Returns whether the model should be auto-populated.
        """
        return self._auto_populate_

    @property
    def auto_generate(self) -> bool:
        """
        Returns whether the model should be auto-generated.
        """
        return self._auto_generate_

    @classmethod
    def generate_config(
        cls,
        config_name: Optional[str] = None,
        config_path: str = None,
        config_root: Optional[str] = None,
        save: bool = True,
    ) -> Dict[str, Any]:
        """
        Saves a HyFI config for itself.

        Args:
            cls (BaseModel): The class to generate a config for.
            config_name (Optional[str]): The name of the config. If not provided, the name of the target will be used.
            config_path (Optional[str]): The path to save the config to (relative to the config root). Defaults to "run".
            config_root (Optional[str]): The root of the config path. If not provided, the global hyfi config directory will be used.
            **kwargs_for_target: Keyword arguments to pass to the target.
        """

        target = f"{cls.__module__}.{cls.__name__}"
        cfg = {
            "_target_": target,
            "_config_name_": getattr(cls._config_name_, "default"),
            "_config_group_": getattr(cls._config_group_, "default"),
        }

        model_fields = {
            key: getattr(value, "default") for key, value in cls.model_fields.items()
        }
        cfg.update(model_fields)
        cfg = cls.sanitized_config(cfg)

        if not save:
            return cfg

        config_name = (
            config_name or getattr(cls._config_name_, "default")
        ) or cls._config_name_
        _save_config(
            config=cfg,
            class_name=cls.__name__,
            config_group=getattr(cls._config_group_, "default"),
            config_name=config_name,
            config_path=config_path,
            config_root=config_root,
        )
        return cfg

    @staticmethod
    def sanitized_config(
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Converts a config to Hydra-supported type if necessary and possible.

        Args:
            config (Dict[str, Any]): The config to sanitize.

        Returns:
            Dict[str, Any]: The sanitized config.
        """
        defaults = []
        sanitized_config = {}
        _config = {}
        for key, value in config.items():
            if hasattr(value, "_config_group_") and hasattr(value, "_config_name_"):
                config_name = (
                    getattr(value, "model_extra", {}).get("_config_name_")
                    or value._config_name_
                )
                if value._config_group_ == key:
                    defaults.append({f"{value._config_group_}": config_name})
                else:
                    defaults.append({f"{value._config_group_}@{key}": config_name})
            else:
                value = sanitized_default_value(value)
                _config[key] = value
        if defaults:
            sanitized_config["defaults"] = defaults
        sanitized_config.update(_config)
        return sanitized_config


def _save_config(
    config: Dict[str, Any],
    class_name: str,
    config_name: str,
    config_group: Optional[str] = None,
    config_root: Optional[str] = None,
    config_path: Optional[str] = None,
):
    filename = f"{config_name}.yaml"
    config_root = config_root or global_hyfi.config_root
    if config_group and config_group.startswith("/"):
        config_group = config_group[1:]
    config_path = config_path or config_group or "test"
    config_path = Path(config_root) / config_path
    config_path.mkdir(parents=True, exist_ok=True)
    config_path /= filename
    Composer.save(config, config_path)
    logger.info("Saved HyFI config for %s to %s", class_name, config_path)


class InnerTestModel(BaseModel):
    _config_name_: str = "__inner__"
    _config_group_: str = "/test"

    name: str = "inner"


class TestModel(BaseModel):
    _config_name_: str = "__test__"
    _config_group_: str = "/test"
    inner: InnerTestModel = InnerTestModel()

    name: str = "test"


__all__ = ["BaseModel", "model_validator"]



"""cached download and extract archive"""
from pathlib import Path

import gdown

from hyfi.cached_path import _cached_path

from .iolibs import IOLIBs
from .logging import LOGGING

logger = LOGGING.getLogger(__name__)


class CACHED:
    @staticmethod
    def cached_path(
        url_or_filename: str,
        extract_archive: bool = False,
        force_extract: bool = False,
        return_parent_dir: bool = False,
        cache_dir: str = "",
        verbose: bool = False,
    ):
        """
        Attempts to cache a file or URL and return the path to the cached file.
        If required libraries 'cached_path' and 'gdown' are not installed, raises an ImportError.

        Args:
            url_or_filename (str): The URL or filename to be cached.
            extract_archive (bool, optional): Whether to extract the file if it's an archive. Defaults to False.
            force_extract (bool, optional): Whether to force extraction even if the destination already exists. Defaults to False.
            return_parent_dir (bool, optional): If True, returns the parent directory of the cached file. Defaults to False.
            cache_dir (str, optional): Directory to store cached files. Defaults to None.
            verbose (bool, optional): Whether to print informative messages during the process. Defaults to False.

        Raises:
            ImportError: If the required libraries 'cached_path' and 'gdown' are not imported.

        Returns:
            str: Path to the cached file or its parent directory, depending on the 'return_parent_dir' parameter.
        """
        if not url_or_filename:
            logger.warning("url_or_filename not provided")
            return None
        if verbose:
            logger.info(
                "caching path: %s, extract_archive: %s, force_extract: %s, cache_dir: %s",
                url_or_filename,
                extract_archive,
                force_extract,
                cache_dir,
            )

        try:
            if url_or_filename.startswith("gd://"):
                _path = IOLIBs.cached_gdown(
                    url_or_filename,
                    verbose=verbose,
                    extract_archive=extract_archive,
                    force_extract=force_extract,
                    cache_dir=cache_dir,
                )
                _path = Path(_path) if isinstance(_path, str) else None
            else:
                if _cached_path is None:
                    raise ImportError(
                        "Error importing required libraries 'cached-path'. "
                        "Please install them using 'pip install cached-path' and try again."
                    )

                if cache_dir:
                    cache_dir = str(Path(cache_dir) / "cached_path")
                else:
                    cache_dir = str(Path.home() / ".hyfi" / ".cache" / "cached_path")

                _path = _cached_path.cached_path(
                    url_or_filename,
                    extract_archive=extract_archive,
                    force_extract=force_extract,
                    cache_dir=cache_dir,
                )

            logger.debug("cached path: %s", _path)

            if _path and _path.is_file():
                _parent_dir = Path(_path).parent
            elif _path and _path.is_dir():
                _parent_dir = Path(_path)
            else:
                logger.warning("Unknown path: %s", _path)
                return None

            return _parent_dir.as_posix() if return_parent_dir else _path
        except Exception as e:
            logger.error(e)
            return None

    @staticmethod
    def cached_gdown(
        url: str,
        verbose: bool = False,
        extract_archive: bool = False,
        force_extract: bool = False,
        cache_dir: str = "",
    ):
        """
        :type url: str
            ex) gd://id:path
        :type verbose: bool
        :type extract_archive: bool
        :type force_extract: bool
        :type cache_dir: str
        :returns: str
        """
        if gdown is None:
            raise ImportError(
                "Error importing required libraries 'gdown'. "
                "Please install them using 'pip install gdown' and try again."
            )

        if verbose:
            logger.info("Downloading %s...", url)
        if cache_dir:
            cache_dir_ = Path(cache_dir) / "gdown"
        else:
            cache_dir_ = Path.home() / ".hyfi" / ".cache" / "gdown"
        cache_dir_.mkdir(parents=True, exist_ok=True)

        gd_prefix = "gd://"
        if url.startswith(gd_prefix):
            url = url[len(gd_prefix) :]
            _url = url.split(":")
            if len(_url) == 2:
                id_, path = _url
            else:
                id_ = _url[0]
                path = id_

            # If we're using the path!c/d/file.txt syntax, handle it here.
            fname = None
            extraction_path = path
            exclamation_index = path.find("!")
            if extract_archive and exclamation_index >= 0:
                extraction_path = path[:exclamation_index]
                fname = path[exclamation_index + 1 :]

            cache_path = cache_dir_ / f".{id_}" / extraction_path
            cache_path.parent.mkdir(parents=True, exist_ok=True)

            cache_path = gdown.cached_download(
                id=id_,
                path=cache_path.as_posix(),
                quiet=not verbose,
            )

            if extract_archive:
                extraction_path, files = IOLIBs.extractall(
                    cache_path, force_extract=force_extract
                )

                if not fname or not files:
                    return extraction_path
                for f in files:
                    if f.endswith(fname):
                        return f
            return cache_path

        else:
            logger.warning("Unknown url: %s", url)
            return None



"""
SimpleEval - (C) 2013-2023 Daniel Fairhead
-------------------------------------

An short, easy to use, safe and reasonably extensible expression evaluator.
Designed for things like in a website where you want to allow the user to
generate a string, or a number from some other input, without allowing full
eval() or other unsafe or needlessly complex linguistics.

-------------------------------------

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

-------------------------------------

Initial idea copied from J.F. Sebastian on Stack Overflow
( http://stackoverflow.com/a/9558001/1973500 ) with
modifications and many improvements.

-------------------------------------
Contributors:
- corro (Robin Baumgartner) (py3k)
- dratchkov (David R) (nested dicts)
- marky1991 (Mark Young) (slicing)
- T045T (Nils Berg) (!=, py3kstr, obj.
- perkinslr (Logan Perkins) (.__globals__ or .func_ breakouts)
- impala2 (Kirill Stepanov) (massive _eval refactor)
- gk (ugik) (Other iterables than str can DOS too, and can be made)
- daveisfera (Dave Johansen) 'not' Boolean op, Pycharm, pep8, various other fixes
- xaled (Khalid Grandi) method chaining correctly, double-eval bugfix.
- EdwardBetts (Edward Betts) spelling correction.
- charlax (Charles-Axel Dein charlax) Makefile and cleanups
- mommothazaz123 (Andrew Zhu) f"string" support, Python 3.8 support
- lubieowoce (Uryga) various potential vulnerabilities
- JCavallo (Jean Cavallo) names dict shouldn't be modified
- Birne94 (Daniel Birnstiel) for fixing leaking generators, star expressions
- patricksurry (Patrick Surry) or should return last value, even if falsy.
- shughes-uk (Samantha Hughes) python w/o 'site' should not fail to import.
- KOLANICH packaging / deployment / setup help & << + >> & other bit ops
- graingert (Thomas Grainger) packaging / deployment / setup help
- bozokopic (Bozo Kopic) Memory leak fix
- daxamin (Dax Amin) Better error for attempting to eval empty string
- smurfix (Matthias Urlichs) Allow clearing functions / operators / etc completely

The name of the module has been changed to SafeEval to reflect the fact that
the purpose of this module is to provide a safe eval() function, not to be a
full expression evaluator.
-------------------------------------
Basic Usage:

>>> s = SafeEval()
>>> s.eval("20 + 30")
50

You can add your own functions easily too:

if file.txt contents is "11"

>>> def get_file():
...     with open("file.txt", 'r') as f:
...         return f.read()

>>> s.functions["get_file"] = get_file
>>> s.eval("int(get_file()) + 31")
42

For more information, see the full package documentation on pypi, or the github
repo.

-----------

If you don't need to re-use the evaluator (with it's names, functions, etc),
then you can use the safe_eval() function:

>>> safe_eval("21 + 19")
40

You can pass names, operators and functions to the safe_eval function as
well:

>>> safe_eval("40 + two", names={"two": 2})
42

"""

import ast
import contextlib
import operator as op
import sys
import warnings
from random import random
from typing import Any, Callable, Dict, Optional

PYTHON3 = sys.version_info[0] == 3
PYTHON35 = PYTHON3 and sys.version_info > (3, 5)

########################################
# Module wide 'globals'

MAX_STRING_LENGTH = 100000
MAX_COMPREHENSION_LENGTH = 10000
MAX_POWER = 4000000  # highest exponent
MAX_SHIFT = 10000  # highest << or >> (lshift / rshift)
MAX_SHIFT_BASE = int(sys.float_info.max)  # highest on left side of << or >>
DISALLOW_PREFIXES = ["_", "func_"]
DISALLOW_METHODS = ["format", "format_map", "mro"]

# Disallow functions:
# This, strictly speaking, is not necessary.  These /should/ never be accessable anyway,
# if DISALLOW_PREFIXES and DISALLOW_METHODS are all right.  This is here to try and help
# people not be stupid.  Allowing these functions opens up all sorts of holes - if any of
# their functionality is required, then please wrap them up in a safe container.  And think
# very hard about it first.  And don't say I didn't warn you.
# builtins is a dict in python >3.6 but a module before
DISALLOW_FUNCTIONS = {type, isinstance, eval, getattr, setattr, repr, compile, open}
if hasattr(__builtins__, "help") or (
    hasattr(__builtins__, "__contains__") and "help" in __builtins__  # type: ignore
):
    # PyInstaller environment doesn't include this module.
    DISALLOW_FUNCTIONS.add(help)


if PYTHON3:
    # exec is not a function in Python2...
    exec("DISALLOW_FUNCTIONS.add(exec)")  # pylint: disable=exec-used


class InvalidExpression(Exception):
    """Generic Exception"""

    pass


class FunctionNotDefined(InvalidExpression):
    """sorry! That function isn't defined!"""

    def __init__(self, func_name, expression):
        self.message = "Function '{0}' not defined," " for expression '{1}'.".format(
            func_name, expression
        )
        setattr(self, "func_name", func_name)  # bypass 2to3 confusion.
        self.expression = expression

        super(InvalidExpression, self).__init__(self.message)


class NameNotDefined(InvalidExpression):
    """a name isn't defined."""

    def __init__(self, name, expression):
        self.name = name
        self.message = "'{0}' is not defined for expression '{1}'".format(
            name, expression
        )
        self.expression = expression

        super(InvalidExpression, self).__init__(self.message)


class AttributeDoesNotExist(InvalidExpression):
    """attribute does not exist"""

    def __init__(self, attr, expression):
        self.message = "Attribute '{0}' does not exist in expression '{1}'".format(
            attr, expression
        )
        self.attr = attr
        self.expression = expression

        super(InvalidExpression, self).__init__(self.message)


class OperatorNotDefined(InvalidExpression):
    """operator does not exist"""

    def __init__(self, attr, expression):
        self.message = "Operator '{0}' does not exist in expression '{1}'".format(
            attr, expression
        )
        self.attr = attr
        self.expression = expression

        super(InvalidExpression, self).__init__(self.message)


class FeatureNotAvailable(InvalidExpression):
    """What you're trying to do is not allowed."""

    pass


class NumberTooHigh(InvalidExpression):
    """Sorry! That number is too high. I don't want to spend the
    next 10 years evaluating this expression!"""

    pass


class IterableTooLong(InvalidExpression):
    """That iterable is **way** too long, baby."""

    pass


class AssignmentAttempted(UserWarning):
    """Assignment not allowed in SafeEval"""

    pass


class MultipleExpressions(UserWarning):
    """Only the first expression parsed will be used"""

    pass


########################################
# Default simple functions to include:


def random_int(top: int) -> int:
    """return a random int below <top>"""

    return int(random() * top)


def safe_power(a, b):  # pylint: disable=invalid-name
    """a limited exponent/to-the-power-of function, for safety reasons"""

    if abs(a) > MAX_POWER or abs(b) > MAX_POWER:
        raise NumberTooHigh("Sorry! I don't want to evaluate {0} ** {1}".format(a, b))
    return a**b


def safe_mult(a, b):  # pylint: disable=invalid-name
    """limit the number of times an iterable can be repeated..."""

    if hasattr(a, "__len__") and b * len(a) > MAX_STRING_LENGTH:
        raise IterableTooLong("Sorry, I will not evalute something that long.")
    if hasattr(b, "__len__") and a * len(b) > MAX_STRING_LENGTH:
        raise IterableTooLong("Sorry, I will not evalute something that long.")

    return a * b


def safe_add(a, b):  # pylint: disable=invalid-name
    """iterable length limit again"""

    if (
        hasattr(a, "__len__")
        and hasattr(b, "__len__")
        and len(a) + len(b) > MAX_STRING_LENGTH
    ):
        raise IterableTooLong(
            "Sorry, adding those two together would" " make something too long."
        )
    return a + b


def safe_rshift(a, b):  # pylint: disable=invalid-name
    """rshift, but with input limits"""
    if abs(b) > MAX_SHIFT or abs(a) > MAX_SHIFT_BASE:
        raise NumberTooHigh("Sorry! I don't want to evaluate {0} >> {1}".format(a, b))
    return a >> b


def safe_lshift(a, b):  # pylint: disable=invalid-name
    """lshift, but with input limits"""
    if abs(b) > MAX_SHIFT or abs(a) > MAX_SHIFT_BASE:
        raise NumberTooHigh("Sorry! I don't want to evaluate {0} << {1}".format(a, b))
    return a << b


########################################
# Defaults for the evaluator:

DEFAULT_OPERATORS = {
    ast.Add: safe_add,
    ast.Sub: op.sub,
    ast.Mult: safe_mult,
    ast.Div: op.truediv,
    ast.FloorDiv: op.floordiv,
    ast.RShift: safe_rshift,
    ast.LShift: safe_lshift,
    ast.Pow: safe_power,
    ast.Mod: op.mod,
    ast.Eq: op.eq,
    ast.NotEq: op.ne,
    ast.Gt: op.gt,
    ast.Lt: op.lt,
    ast.GtE: op.ge,
    ast.LtE: op.le,
    ast.Not: op.not_,
    ast.USub: op.neg,
    ast.UAdd: op.pos,
    ast.BitXor: op.xor,
    ast.BitOr: op.or_,
    ast.BitAnd: op.and_,
    ast.Invert: op.invert,
    ast.In: lambda x, y: op.contains(y, x),
    ast.NotIn: lambda x, y: not op.contains(y, x),
    ast.Is: lambda x, y: x is y,
    ast.IsNot: lambda x, y: x is not y,
}

DEFAULT_FUNCTIONS = {
    "rand": random,
    "randint": random_int,
    "int": int,
    "float": float,
    "str": str,
}

DEFAULT_NAMES = {"True": True, "False": False, "None": None}

ATTR_INDEX_FALLBACK = True


########################################
# And the actual evaluator:


class SafeEval(object):  # pylint: disable=too-few-public-methods
    """A very simple expression parser.
    >>> s = SafeEval()
    >>> s.eval("20 + 30 - ( 10 * 5)")
    0
    """

    expr = ""

    def __init__(
        self,
        operators: Optional[Dict[Any, Callable]] = None,
        functions: Optional[Dict[str, Callable]] = None,
        names: Optional[Dict[str, Any]] = None,
    ):
        """
        Create the evaluator instance.  Set up valid operators (+,-, etc)
        functions (add, random, get_val, whatever) and names."""

        if operators is None:
            operators = DEFAULT_OPERATORS.copy()
        if functions is None:
            functions = DEFAULT_FUNCTIONS.copy()
        if names is None:
            names = DEFAULT_NAMES.copy()

        self.operators = operators
        self.functions = functions
        self.names = names

        self.nodes = {
            ast.Expr: self._eval_expr,
            ast.Assign: self._eval_assign,
            ast.AugAssign: self._eval_aug_assign,
            ast.Import: self._eval_import,
            ast.Num: self._eval_num,
            ast.Str: self._eval_str,
            ast.Name: self._eval_name,
            ast.UnaryOp: self._eval_unaryop,
            ast.BinOp: self._eval_binop,
            ast.BoolOp: self._eval_boolop,
            ast.Compare: self._eval_compare,
            ast.IfExp: self._eval_ifexp,
            ast.Call: self._eval_call,
            ast.keyword: self._eval_keyword,
            ast.Subscript: self._eval_subscript,
            ast.Attribute: self._eval_attribute,
            ast.Index: self._eval_index,
            ast.Slice: self._eval_slice,
        }

        # py3k stuff:
        if hasattr(ast, "NameConstant"):
            self.nodes[ast.NameConstant] = self._eval_constant

        # py3.6, f-strings
        if hasattr(ast, "JoinedStr"):
            self.nodes[ast.JoinedStr] = self._eval_joinedstr  # f-string
            self.nodes[
                ast.FormattedValue
            ] = self._eval_formattedvalue  # formatted value in f-string

        # py3.8 uses ast.Constant instead of ast.Num, ast.Str, ast.NameConstant
        if hasattr(ast, "Constant"):
            self.nodes[ast.Constant] = self._eval_constant

        # Defaults:

        self.ATTR_INDEX_FALLBACK = ATTR_INDEX_FALLBACK

        # Check for forbidden functions:

        for f in self.functions.values():
            if f in DISALLOW_FUNCTIONS:
                raise FeatureNotAvailable(f"This function {f} is a really bad idea.")

    def __del__(self):
        self.nodes = None

    @staticmethod
    def parse(expr):
        """parse an expression into a node tree"""

        parsed = ast.parse(expr.strip())

        if not parsed.body:
            raise InvalidExpression("Sorry, cannot evaluate empty string")
        if len(parsed.body) > 1:
            warnings.warn(
                f"'{expr}' contains multiple expressions. Only the first will be used.",
                MultipleExpressions,
            )
        return parsed.body[0]

    def eval(self, expr, previously_parsed=None):
        """evaluate an expresssion, using the operators, functions and
        names previously set up."""

        # set a copy of the expression aside, so we can give nice errors...
        self.expr = expr

        return self._eval(previously_parsed or self.parse(expr))

    def _eval(self, node):
        """The internal evaluator used on each node in the parsed tree."""

        try:
            handler = self.nodes[type(node)]
        except KeyError as e:
            raise FeatureNotAvailable(
                "Sorry, {0} is not available in this "
                "evaluator".format(type(node).__name__)
            ) from e

        return handler(node)

    def _eval_expr(self, node):
        return self._eval(node.value)

    def _eval_assign(self, node):
        warnings.warn(
            f"Assignment ({self.expr}) attempted, but this is ignored",
            AssignmentAttempted,
        )
        return self._eval(node.value)

    def _eval_aug_assign(self, node):
        warnings.warn(
            f"Assignment ({self.expr}) attempted, but this is ignored",
            AssignmentAttempted,
        )
        return self._eval(node.value)

    @staticmethod
    def _eval_import(node):
        raise FeatureNotAvailable("Sorry, 'import' is not allowed.")

    @staticmethod
    def _eval_num(node):
        return node.n

    @staticmethod
    def _eval_str(node):
        if len(node.s) > MAX_STRING_LENGTH:
            raise IterableTooLong(
                "String Literal in statement is too long!"
                " ({0}, when {1} is max)".format(len(node.s), MAX_STRING_LENGTH)
            )
        return node.s

    @staticmethod
    def _eval_constant(node):
        if hasattr(node.value, "__len__") and len(node.value) > MAX_STRING_LENGTH:
            raise IterableTooLong(
                "Literal in statement is too long!"
                " ({0}, when {1} is max)".format(len(node.value), MAX_STRING_LENGTH)
            )
        return node.value

    def _eval_unaryop(self, node):
        try:
            operator = self.operators[type(node.op)]
        except KeyError as e:
            raise OperatorNotDefined(node.op, self.expr) from e
        return operator(self._eval(node.operand))

    def _eval_binop(self, node):
        try:
            operator = self.operators[type(node.op)]
        except KeyError as e:
            raise OperatorNotDefined(node.op, self.expr) from e
        return operator(self._eval(node.left), self._eval(node.right))

    def _eval_boolop(self, node):
        to_return = False
        for value in node.values:
            if isinstance(node.op, ast.And):
                to_return = self._eval(value)
                if not to_return:
                    break
            elif isinstance(node.op, ast.Or):
                to_return = self._eval(value)
                if to_return:
                    break
        return to_return

    def _eval_compare(self, node):
        right = self._eval(node.left)
        to_return = True
        for operation, comp in zip(node.ops, node.comparators):
            if not to_return:
                break
            left = right
            right = self._eval(comp)
            to_return = self.operators[type(operation)](left, right)
        return to_return

    def _eval_ifexp(self, node):
        return (
            self._eval(node.body) if self._eval(node.test) else self._eval(node.orelse)
        )

    def _eval_call(self, node):
        if isinstance(node.func, ast.Attribute):
            func = self._eval(node.func)
        else:
            try:
                func = self.functions[node.func.id]
            except KeyError as e:
                raise FunctionNotDefined(node.func.id, self.expr) from e
            except AttributeError as e:
                raise FeatureNotAvailable("Lambda Functions not implemented") from e

            if func in DISALLOW_FUNCTIONS:
                raise FeatureNotAvailable("This function is forbidden")

        return func(
            *(self._eval(a) for a in node.args),
            **dict(self._eval(k) for k in node.keywords),
        )

    def _eval_keyword(self, node):
        return node.arg, self._eval(node.value)

    def _eval_name(self, node):
        try:
            # This happens at least for slicing
            # This is a safe thing to do because it is impossible
            # that there is a true exression assigning to none
            # (the compiler rejects it, so you can't even
            # pass that to ast.parse)
            if hasattr(self.names, "__getitem__"):
                return self.names[node.id]
            if callable(self.names):
                return self.names(node)
            raise InvalidExpression(
                'Trying to use name (variable) "{0}"'
                ' when no "names" defined for'
                " evaluator".format(node.id)
            )

        except KeyError as e:
            if node.id in self.functions:
                return self.functions[node.id]

            raise NameNotDefined(node.id, self.expr) from e

    def _eval_subscript(self, node):
        container = self._eval(node.value)
        key = self._eval(node.slice)
        try:
            return container[key]
        except KeyError as e:
            raise InvalidExpression(f"Invalid key: {key}") from e

    def _eval_attribute(self, node):
        attr = node.attr
        if any(attr.startswith(prefix) for prefix in DISALLOW_PREFIXES):
            raise FeatureNotAvailable(
                f"Sorry, access to __attributes or func_ attributes is not available. ({attr})"
            )
        if attr in DISALLOW_METHODS:
            raise FeatureNotAvailable(f"Sorry, this method is not available. ({attr})")

        node_evaluated = self._eval(node.value)

        try:
            return getattr(node_evaluated, attr)
        except (AttributeError, TypeError) as e:
            if self.ATTR_INDEX_FALLBACK:
                with contextlib.suppress(KeyError, TypeError):
                    return node_evaluated[attr]
            raise AttributeDoesNotExist(attr, self.expr) from e

    def _eval_index(self, node):
        return self._eval(node.value)

    def _eval_slice(self, node):
        lower = upper = step = None
        if node.lower is not None:
            lower = self._eval(node.lower)
        if node.upper is not None:
            upper = self._eval(node.upper)
        if node.step is not None:
            step = self._eval(node.step)
        return slice(lower, upper, step)

    def _eval_joinedstr(self, node):
        length = 0
        evaluated_values = []
        for n in node.values:
            val = str(self._eval(n))
            if len(val) + length > MAX_STRING_LENGTH:
                raise IterableTooLong("Sorry, I will not evaluate something this long.")
            evaluated_values.append(val)
        return "".join(evaluated_values)

    def _eval_formattedvalue(self, node):
        if node.format_spec:
            fmt = "{:" + self._eval(node.format_spec) + "}"
            return fmt.format(self._eval(node.value))
        return self._eval(node.value)


class EvalWithCompoundTypes(SafeEval):
    """
    SafeEval with additional Compound Types, and their respective
    function editions. (list, tuple, dict, set).
    """

    _max_count = 0

    def __init__(self, operators=None, functions=None, names=None):
        super(EvalWithCompoundTypes, self).__init__(operators, functions, names)

        self.functions.update(list=list, tuple=tuple, dict=dict, set=set)

        self.nodes.update(
            {
                ast.Dict: self._eval_dict,
                ast.Tuple: self._eval_tuple,
                ast.List: self._eval_list,
                ast.Set: self._eval_set,
                ast.ListComp: self._eval_comprehension,
                ast.GeneratorExp: self._eval_comprehension,
            }
        )

    def eval(self, expr, previously_parsed=None):
        # reset _max_count for each eval run
        self._max_count = 0
        return super(EvalWithCompoundTypes, self).eval(expr, previously_parsed)

    def _eval_dict(self, node):
        result = {}

        for key, value in zip(node.keys, node.values):
            if PYTHON35 and key is None:
                # "{**x}" gets parsed as a key-value pair of (None, Name(x))
                result.update(self._eval(value))
            else:
                result[self._eval(key)] = self._eval(value)

        return result

    def _eval_list(self, node):
        result = []

        for item in node.elts:
            if PYTHON3 and isinstance(item, ast.Starred):
                result.extend(self._eval(item.value))
            else:
                result.append(self._eval(item))

        return result

    def _eval_tuple(self, node):
        return tuple(self._eval(x) for x in node.elts)

    def _eval_set(self, node):
        return {self._eval(x) for x in node.elts}

    def _eval_comprehension(self, node):
        to_return = []

        extra_names = {}

        previous_name_evaller = self.nodes[ast.Name]

        def eval_names_extra(node):
            """
            Here we hide our extra scope for within this comprehension
            """
            if node.id in extra_names:
                return extra_names[node.id]
            return previous_name_evaller(node)

        self.nodes.update({ast.Name: eval_names_extra})

        def recurse_targets(target, value):
            """
                Recursively (enter, (into, (nested, name), unpacking)) = \
                             and, (assign, (values, to), each
            """
            if isinstance(target, ast.Name):
                extra_names[target.id] = value
            else:
                for t, v in zip(target.elts, value):
                    recurse_targets(t, v)

        def do_generator(gi=0):
            g = node.generators[gi]
            for i in self._eval(g.iter):
                self._max_count += 1

                if self._max_count > MAX_COMPREHENSION_LENGTH:
                    raise IterableTooLong("Comprehension generates too many elements")
                recurse_targets(g.target, i)
                if all(self._eval(iff) for iff in g.ifs):
                    if len(node.generators) > gi + 1:
                        do_generator(gi + 1)
                    else:
                        to_return.append(self._eval(node.elt))

        try:
            do_generator()
        finally:
            self.nodes.update({ast.Name: previous_name_evaller})

        return to_return


class SAFEEVAL:
    @staticmethod
    def safe_eval(
        expr: str,
        operators=None,
        functions: Optional[Dict[str, Callable]] = None,
        names: Optional[Dict[str, Any]] = None,
    ):
        """Simply evaluate an expresssion"""
        s = SafeEval(operators=operators, functions=functions, names=names)
        return s.eval(expr)



"""Motion image processing functions."""
import os
import subprocess
from pathlib import Path
from typing import Optional

from hyfi.utils.iolibs import IOLIBs
from hyfi.utils.logging import LOGGING
from hyfi.utils.notebooks import NBs

from .utils import GUTILs

logger = LOGGING.getLogger(__name__)


class MOTION:
    @staticmethod
    def make_gif(
        image_filepaths=None,
        filename_patterns: Optional[str] = None,
        base_dir: Optional[str] = None,
        output_file: Optional[str] = None,
        duration: int = 100,
        loop: int = 0,
        width: int = 0,
        optimize: bool = True,
        quality: int = 50,
        display_to_notebook: bool = False,
        force: bool = False,
        **kwargs,
    ) -> Optional[str]:
        """
        Create a GIF from a list of images or a list of filenames.
        """
        output_file = output_file or ""
        if os.path.exists(output_file) and not force:
            logger.info("Skipping GIF creation, already exists: %s", output_file)
            logger.info("If you want to re-create the GIF, set force=True")
        else:
            if image_filepaths is None and filename_patterns:
                image_filepaths = sorted(
                    IOLIBs.get_filepaths(filename_patterns, base_dir=base_dir)
                )
            if not image_filepaths:
                logger.warning("no images found")
                return
            if frames := GUTILs.load_images(image_filepaths):
                frame_one = frames[0]
                frame_one.save(
                    output_file,
                    format="GIF",
                    append_images=frames,
                    save_all=True,
                    duration=duration,
                    loop=loop,
                    optimize=optimize,
                    quality=quality,
                )
                logger.info("Saved GIF to %s", output_file)
            else:
                logger.warning("No frames found for %s", filename_patterns)

        if display_to_notebook and os.path.exists(output_file):
            NBs.display_image(data=IOLIBs.read(output_file), width=width)

        return output_file

    @staticmethod
    def extract_frames(
        input_video_file: str,
        every_nth_frame: int,
        ouput_dir: str,
        frame_filename_pattern: str = "%04d.jpg",
        ffmpeg_path: str = "/usr/bin/ffmpeg",
    ):
        """
        Extract frames from a video.

        Args:
            input_video_file (str): Path to the video.
            every_nth_frame (int): Extract every nth frame.
            ouput_dir (str): Path to the output directory.
            frame_filename_pattern (str, optional): Frame filename pattern. Defaults to "%04d.jpg".
        """
        logger.info(
            "Exporting Video Frames (every %sth frame) to %s",
            every_nth_frame,
            ouput_dir,
        )
        try:
            for f in Path(f"{ouput_dir}").glob("*.jpg"):
                f.unlink()
        except FileNotFoundError:
            logger.info("No video frames found in %s", ouput_dir)
        vf = f"select=not(mod(n\\,{every_nth_frame}))"

        if not os.path.exists(ffmpeg_path):
            ffmpeg_path = "ffmpeg"
        if os.path.exists(input_video_file):
            subprocess.run(
                [
                    ffmpeg_path,
                    "-i",
                    f"{input_video_file}",
                    "-vf",
                    f"{vf}",
                    "-vsync",
                    "vfr",
                    "-q:v",
                    "2",
                    "-loglevel",
                    "error",
                    "-stats",
                    f"{ouput_dir}/{frame_filename_pattern}",
                ],
                stdout=subprocess.PIPE,
            ).stdout.decode("utf-8")
        else:
            logger.warning(
                "WARNING!\n\nVideo not found: %s.\nPlease check your video path.",
                input_video_file,
            )

    @staticmethod
    def create_video(
        input_image_dir: str,
        video_output_file: str,
        input_url: str,
        fps: int,
        start_number: int,
        vframes: int,
        force: bool = False,
    ) -> Optional[str]:
        """
        Create a video from a list of images.

        Args:
            input_image_dir (str): Base directory.
            video_path (str): Path to the video.
            input_url (str): Input URL.
            fps (int): Frames per second.
            start_number (int): Start number.
            vframes (int): Number of frames.
            force (bool, optional): Force video creation. Defaults to False.

        Raises:
            FileNotFoundError: If the video is not found.

        """

        logger.info("Creating video from %s", input_url)
        if os.path.exists(video_output_file) and not force:
            logger.info(
                "Skipping video creation, already exists: %s", video_output_file
            )
            logger.info("If you want to re-create the video, set force=True")
            return video_output_file

        cmd = [
            "ffmpeg",
            "-y",
            "-vcodec",
            "png",
            "-r",
            str(fps),
            "-start_number",
            str(start_number),
            "-i",
            input_url,
            "-frames:v",
            str(vframes),
            "-c:v",
            "libx264",
            "-vf",
            f"fps={fps}",
            "-pix_fmt",
            "yuv420p",
            "-crf",
            "17",
            "-preset",
            "veryslow",
            video_output_file,
        ]

        process = subprocess.Popen(
            cmd,
            cwd=f"{input_image_dir}",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _, stderr = process.communicate()
        if process.returncode != 0:
            logger.error(stderr)
            raise RuntimeError(stderr)
        else:
            logger.info("Video created successfully and saved to %s", video_output_file)

        return video_output_file



from pydantic import (
    ConfigDict,
    Field,
    FieldValidationInfo,
    PrivateAttr,
    SecretStr,
    field_validator,
    model_validator,
)
from pydantic_settings import PydanticBaseSettingsSource, SettingsConfigDict

from .composer import Composer, SpecialKeys
from .config import BaseConfig
from .generator import GENERATOR, PipeTargetTypes
from .model import BaseModel
from .settings import BaseSettings

__all__ = [
    "BaseConfig",
    "BaseModel",
    "BaseSettings",
    "Composer",
    "ConfigDict",
    "field_validator",
    "Field",
    "FieldValidationInfo",
    "GENERATOR",
    "model_validator",
    "PipeTargetTypes",
    "PrivateAttr",
    "PydanticBaseSettingsSource",
    "SecretStr",
    "SettingsConfigDict",
    "SpecialKeys",
]



import datetime
import os
from contextlib import contextmanager
from functools import partial
from timeit import default_timer

from hyfi.utils.funcs import FUNCs


def _elapser_timer(start):
    """A function that returns the elapsed time"""
    return default_timer() - start


def _elapser(start, end):
    """A function that returns the elapsed time"""
    return end - start


@contextmanager
def elapsed_timer(format_time=False):
    """A context manager that yields a function that returns the elapsed time"""
    start = default_timer()
    # elapser = lambda: default_timer() - start
    elapser = partial(_elapser_timer, start)
    yield lambda: str(
        datetime.timedelta(seconds=elapser())
    ) if format_time else elapser()
    end = default_timer()
    # elapser = lambda: end - start
    elapser = partial(_elapser, start, end)


@contextmanager
def change_directory(
    directory: str,
    ensure_directory: bool = True,
):
    """Change directory and change back to original directory"""
    original = os.path.abspath(os.getcwd())

    FUNCs.fancy_print(f" Change directory to {directory}")
    if ensure_directory:
        os.makedirs(directory, exist_ok=True)
    os.chdir(directory)
    try:
        yield

    except Exception as e:
        FUNCs.fancy_print(f" Exception: {e}")
        raise e

    finally:
        FUNCs.fancy_print(f" Change directory back to {original}")
        os.chdir(original)



import os
import tarfile
from hashlib import sha256
from pathlib import Path
from typing import List, Optional, Tuple
from urllib.parse import urlparse

from .common import PathOrStr, get_cache_dir
from .meta import Meta


def resource_to_filename(resource: PathOrStr, etag: Optional[str] = None) -> str:
    """
    Convert a ``resource`` into a hashed filename in a repeatable way.
    If ``etag`` is specified, append its hash to the resources', delimited
    by a period.

    THis is essentially the inverse of :func:`filename_to_url()`.
    """
    resource_bytes = str(resource).encode("utf-8")
    resource_hash = sha256(resource_bytes)
    filename = resource_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += f".{etag_hash.hexdigest()}"

    return filename


def filename_to_url(
    filename: str, cache_dir: Optional[PathOrStr] = None
) -> Tuple[str, Optional[str]]:
    """
    Return the URL and etag (which may be ``None``) stored for ``filename``.
    Raises :exc:`FileNotFoundError` if ``filename`` or its stored metadata do not exist.

    This is essentially the inverse of :func:`resource_to_filename()`.
    """
    cache_dir = cache_dir or get_cache_dir()
    cache_path = os.path.join(cache_dir, filename)
    if not os.path.exists(cache_path):
        raise FileNotFoundError(f"file {cache_path} not found")

    meta_path = f"{cache_path}.json"
    if not os.path.exists(meta_path):
        raise FileNotFoundError(f"file {meta_path} not found")

    metadata = Meta.from_path(meta_path)
    return metadata.resource, metadata.etag


def find_latest_cached(
    url: str, cache_dir: Optional[PathOrStr] = None
) -> Optional[Path]:
    """
    Get the path to the latest cached version of a given resource.
    """
    cache_dir = Path(cache_dir or get_cache_dir())
    filename = resource_to_filename(url)
    candidates: List[Tuple[Path, float]] = []
    for path in cache_dir.glob(f"{filename}*"):
        print(path, path.suffix, path.name)
        if path.suffix in {".json", ".lock"} or path.name.endswith("-extracted"):
            continue
        mtime = path.stat().st_mtime
        candidates.append((path, mtime))
    # Sort candidates by modification time, newest first.
    candidates.sort(key=lambda x: x[1], reverse=True)
    return candidates[0][0] if candidates else None


def check_tarfile(tar_file: tarfile.TarFile):
    """Tar files can contain files outside of the extraction directory, or symlinks that point
    outside the extraction directory. We also don't want any block devices fifos, or other
    weird file types extracted. This checks for those issues and throws an exception if there
    is a problem."""
    base_path = os.path.join("tmp", "pathtest")
    base_path = os.path.normpath(base_path)

    def normalize_path(path: str) -> str:
        path = path.rstrip("/")
        path = path.replace("/", os.sep)
        path = os.path.join(base_path, path)
        path = os.path.normpath(path)
        return path

    for tarinfo in tar_file:
        if not (
            tarinfo.isreg()
            or tarinfo.isdir()
            or tarinfo.isfile()
            or tarinfo.islnk()
            or tarinfo.issym()
        ):
            raise ValueError(
                f"Tar file {str(tar_file.name)} contains invalid member {tarinfo.name}."
            )

        target_path = normalize_path(tarinfo.name)
        if os.path.commonprefix([base_path, target_path]) != base_path:
            raise ValueError(
                f"Tar file {str(tar_file.name)} is trying to create a file outside of its extraction directory."
            )

        if tarinfo.islnk() or tarinfo.issym():
            target_path = normalize_path(tarinfo.linkname)
            if os.path.commonprefix([base_path, target_path]) != base_path:
                raise ValueError(
                    f"Tar file {str(tar_file.name)} is trying to link to a file "
                    "outside of its extraction directory."
                )


def is_url_or_existing_file(url_or_filename: PathOrStr) -> bool:
    """
    Given something that might be a URL or local path,
    determine if it's actually a url or the path to an existing file.
    """
    if url_or_filename is None:
        return False

    from .schemes import get_supported_schemes

    url_or_filename = os.path.expanduser(str(url_or_filename))
    parsed = urlparse(url_or_filename)
    return parsed.scheme in get_supported_schemes() or os.path.exists(url_or_filename)


def _lock_file_path(cache_path: Path) -> Path:
    return cache_path.parent / f"{cache_path.name}.lock"


def _meta_file_path(cache_path: Path) -> Path:
    return cache_path.parent / f"{cache_path.name}.json"



"""
    HiFY Base Config
"""
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from omegaconf import DictConfig

from hyfi.utils.logging import LOGGING

from .composer import Composer
from .model import BaseModel

logger = LOGGING.getLogger(__name__)


class BaseConfig(BaseModel):
    """
    Base class for all config classes.
    """

    _config_name_: str = "__init__"
    _auto_populate_: bool = True

    verbose: bool = False

    _init_args_: Dict[str, Any] = {}
    _property_set_methods_: Dict[str, str] = {}
    _subconfigs_: Dict[str, Any] = {}

    def __init__(self, **config_kwargs):
        logger.debug(
            "init %s with %s args", self.__class__.__name__, len(config_kwargs)
        )
        super().__init__(**config_kwargs)
        # self.initialize_subconfigs(config_kwargs)
        self._init_args_ = config_kwargs.copy()

    def __setattr__(self, key, val):
        """
        Overrides the default __setattr__ method to allow for custom property set methods.

        Args:
            key (str): The name of the attribute to set.
            val (Any): The value to set the attribute to.
        """
        if method := self._property_set_methods_.get(key):  # type: ignore
            logger.debug(
                "Setting %s to %s",
                key,
                val if isinstance(val, (str, int)) else type(val),
            )
            getattr(self, method)(val)
        super().__setattr__(key, val)

    def export_config(
        self,
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> Dict[str, Any]:
        """
        Export the configuration to a dictionary.

        Args:
            exclude (Optional[Union[str, List[str], Set[str], None]]): Keys to exclude from the saved configuration.
                Defaults to None.
            exclude_none (bool): Whether to exclude keys with None values. Defaults to True.
            only_include (Optional[Union[str, List[str], Set[str], None]]): Keys to include in the saved configuration.
                Defaults to None.

        Returns:
            Dict[str, Any]: The configuration dictionary.
        """
        if not exclude:
            exclude = self._exclude_  # type: ignore
        if isinstance(exclude, str):
            exclude = [exclude]
        if exclude is None:
            exclude = []
        if isinstance(only_include, str):
            only_include = [only_include]
        if only_include is None:
            only_include = []

        config = self.model_dump(exclude=exclude, exclude_none=exclude_none)  # type: ignore
        if only_include:
            config = {key: config[key] for key in only_include if key in config}

        return config

    def save_config(
        self,
        filepath: Union[str, Path],
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> str:
        """
        Save the batch configuration to file.

        Args:
            filepath ([Union[str, Path]): The filepath to save the configuration to.
            exclude (Optional[Union[str, List[str], Set[str], None]]): Keys to exclude from the saved configuration.
                Defaults to None.
            exclude_none (bool): Whether to exclude keys with None values. Defaults to True.
            only_include (Optional[Union[str, List[str], Set[str], None]]): Keys to include in the saved configuration.
                Defaults to None.

        Returns:
            str: The filename of the saved configuration.
        """
        logger.info("Saving config to %s", filepath)

        config_to_save = self.export_config(
            exclude=exclude, exclude_none=exclude_none, only_include=only_include
        )

        Composer.save(config_to_save, filepath)
        return str(filepath)

    def save_config_as_json(
        self,
        filepath: Union[str, Path],
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> str:
        def dumper(obj):
            return Composer.to_dict(obj) if isinstance(obj, DictConfig) else str(obj)

        config_to_save = self.export_config(
            exclude=exclude, exclude_none=exclude_none, only_include=only_include
        )
        logger.info("Saving config to %s", filepath)
        Composer.save_json(config_to_save, filepath, default=dumper)
        return str(filepath)

    def print_config(
        self,
    ):
        Composer.print(self.model_dump())



"""
Dataset transformation functions. Concatenate, merge, join, etc.
"""
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

import datasets as hfds
import pandas as pd
from datasets.arrow_dataset import Dataset
from datasets.info import DatasetInfo
from datasets.splits import NamedSplit

from hyfi.utils.logging import LOGGING

from .load import DSLoad
from .types import DatasetType

logger = LOGGING.getLogger(__name__)


class DSCombine:
    @staticmethod
    def concatenate_data(
        data: Union[Dict[str, pd.DataFrame], Sequence[pd.DataFrame], List[DatasetType]],
        columns: Optional[Sequence[str]] = None,
        add_split_key_column: bool = False,
        added_column_name: str = "_name_",
        ignore_index: bool = True,
        axis: int = 0,
        split: Optional[str] = None,
        verbose: bool = False,
        **kwargs,
    ) -> Union[pd.DataFrame, DatasetType]:
        # if data is a list of datasets, concatenate them
        if isinstance(data, list) and isinstance(data[0], Dataset):
            return DSCombine.concatenate_datasets(
                data,
                axis=axis,
                split=split,
                **kwargs,
            )
        else:
            return DSCombine.concatenate_dataframes(
                data,
                columns=columns,
                add_split_key_column=add_split_key_column,
                added_column_name=added_column_name,
                ignore_index=ignore_index,
                axis=axis,
                verbose=verbose,
                **kwargs,
            )

    @staticmethod
    def concatenate_dataframes(
        data: Union[Dict[str, pd.DataFrame], Sequence[pd.DataFrame]],
        columns: Optional[Sequence[str]] = None,
        add_split_key_column: bool = False,
        added_column_name: str = "_name_",
        ignore_index: bool = True,
        axis: int = 0,
        verbose: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Concatenate dataframes"""
        if isinstance(data, dict):
            dfs = []
            for split in data:
                df = data[split]
                if add_split_key_column:
                    df[added_column_name] = split
                dfs.append(df)
            data = dfs
        if add_split_key_column:
            columns.append(added_column_name)
        if isinstance(data, list):
            if isinstance(columns, list):
                _columns = [c for c in columns if c in df.columns]
                data = [df[_columns] for df in data]
            if verbose:
                logger.info("Concatenating %s dataframes", len(data))
            if len(data) > 0:
                return pd.concat(data, ignore_index=ignore_index, axis=axis)
            else:
                raise ValueError("No dataframes to concatenate")
        else:
            logger.warning("Warning: data is not a dict")
            return data

    @staticmethod
    def concatenate_datasets(
        dsets: List[DatasetType],
        info: Optional[DatasetInfo] = None,
        split: Optional[NamedSplit] = None,
        axis: int = 0,
    ) -> DatasetType:
        """
        Converts a list of [`Dataset`] with the same schema into a single [`Dataset`].

        Args:
            dsets (`List[datasets.Dataset]`):
                List of Datasets to concatenate.
            info (`DatasetInfo`, *optional*):
                Dataset information, like description, citation, etc.
            split (`NamedSplit`, *optional*):
                Name of the dataset split.
            axis (`{0, 1}`, defaults to `0`):
                Axis to concatenate over, where `0` means over rows (vertically) and `1` means over columns
                (horizontally).

                <Added version="1.6.0"/>

        Example:

        ```py
        >>> ds3 = concatenate_datasets([ds1, ds2])
        ```
        """
        return hfds.concatenate_datasets(
            dsets=dsets,
            info=info,
            split=split,
            axis=axis,
        )

    @staticmethod
    def merge_dataframes(
        data: pd.DataFrame,
        right: Union[str, Path, pd.DataFrame],
        how: str = "inner",
        on: Optional[Union[str, List[str]]] = None,
        left_on: Optional[Union[str, List[str]]] = None,
        right_on: Optional[Union[str, List[str]]] = None,
        left_index: bool = False,
        right_index: bool = False,
        sort: bool = False,
        suffixes: Tuple[str, str] = ("_x", "_y"),
        copy: bool = True,
        indicator: bool = False,
        validate: Optional[str] = None,
        verbose: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Merge two dataframes

        Args:
            left (pd.DataFrame): left dataframe
            right (pd.DataFrame): right dataframe
            how (str, optional): how to merge. Defaults to "inner".
            on (Optional[Union[str, List[str]]], optional): column(s) to merge on. Defaults to None.
            left_on (Optional[Union[str, List[str]]], optional): column(s) to merge on in left dataframe. Defaults to None.
            right_on (Optional[Union[str, List[str]]], optional): column(s) to merge on in right dataframe. Defaults to None.
            left_index (bool, optional): merge on index in left dataframe. Defaults to False.
            right_index (bool, optional): merge on index in right dataframe. Defaults to False.
            sort (bool, optional): sort merged dataframe. Defaults to False.
            suffixes (Optional[Sequence[str]], optional): suffixes to add to column names. Defaults to None.
            copy (bool, optional): copy dataframes. Defaults to True.
            indicator (bool, optional): add indicator column. Defaults to False.
            validate (Optional[str], optional): validation method. Defaults to None.
            verbose (bool, optional): verbose logging. Defaults to False.

        Returns:
            pd.DataFrame: merged dataframe

        Example:
            >>> df1 = pd.DataFrame({"id": [1, 2, 3], "a": [1, 2, 3]})
            >>> df2 = pd.DataFrame({"id": [1, 2, 4], "b": [4, 5, 6]})
            >>> df = DSCombine.merge_dataframes(df1, df2, on="id")
            >>> df
                id  a  b
            0   1  1  4
            1   2  2  5

        """
        if verbose:
            logger.info("Merging dataframes")
        if isinstance(right, (str, Path)):
            right = DSLoad.load_dataframe(right)
        return data.merge(
            right,
            how=how,
            on=on,
            left_on=left_on,
            right_on=right_on,
            left_index=left_index,
            right_index=right_index,
            sort=sort,
            suffixes=suffixes,
            copy=copy,
            indicator=indicator,
            validate=validate,
            **kwargs,
        )



from .core import GlobalHyFIResolver
from .core import __global_hyfi__ as global_hyfi
from .core import _batcher_instance_ as global_batcher

__all__ = [
    "global_hyfi",
    "GlobalHyFIResolver",
    "global_batcher",
]



"""
    Configuration for function calls.
"""
from pathlib import Path
from typing import Any, Dict, Optional

from hyfi.composer import BaseModel, Composer
from hyfi.core import global_hyfi
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class RunConfig(BaseModel):
    """Run Configuration"""

    _config_group_: str = "/run"
    _auto_populate_ = True
    _target_: Optional[str] = None
    _partial_: bool = True

    @property
    def config(self) -> Dict[str, Any]:
        return {
            k: v
            for k, v in self.model_extra.items()
            if not k.startswith("_") or k in ["_target_", "_partial_"]
        }

    @property
    def kwargs(self) -> Dict[str, Any]:
        return {k: v for k, v in self.model_extra.items() if not k.startswith("_")}

    @property
    def target(self) -> str:
        return self._target_

    def generate_config(
        self,
        config_name: Optional[str] = None,
        config_path: str = None,
        config_root: Optional[str] = None,
        save: bool = True,
    ) -> Dict[str, Any]:
        """
        Saves a HyFI config for itself.

        Args:
            config_name (Optional[str]): The name of the config. If not provided, the name of the target will be used.
            config_path (Optional[str]): The path to save the config to (relative to the config root). Defaults to "run".
            config_root (Optional[str]): The root of the config path. If not provided, the global hyfi config directory will be used.
            **kwargs_for_target: Keyword arguments to pass to the target.
        """
        # if self.target is None:
        #     raise ValueError("Cannot generate config for RunConfig without a target.")

        cfg = self.sanitized_config(self.config)

        if not save:
            return cfg

        config_name = config_name or self._config_name_ or self._target_.split(".")[-1]
        filename = f"{config_name}.yaml"
        config_root = config_root or global_hyfi.config_root
        _config_group_ = self._config_group_
        if _config_group_ and _config_group_.startswith("/"):
            _config_group_ = _config_group_[1:]
        config_path = config_path or _config_group_ or "test"
        config_path = Path(config_root) / config_path
        config_path.mkdir(parents=True, exist_ok=True)
        config_path /= filename
        Composer.save(cfg, config_path)
        logger.info("Saved HyFI config for %s to %s.", self._target_, config_path)
        return cfg



from .collage import COLLAGE, Collage
from .motion import MOTION
from .utils import GUTILs


class GRAPHICs(COLLAGE, MOTION, GUTILs):
    def __init__(self):
        pass


__all__ = ["GRAPHICs", "COLLAGE", "MOTION", "GUTILs", "Collage"]



"""
    HyFI Core Module
"""
import os
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Union

from pydantic import BaseModel

from hyfi.utils.logging import LOGGING
from hyfi.utils.packages import PKGs

logger = LOGGING.getLogger(__name__)

__hydra_version_base__: str = "1.2"
__hydra_default_config_group_value__: str = "__init__"
__hyfi_name__: str = "HyFI"
__hyfi_config_dirname__: str = "conf"
__hyfi_config_name__: str = "config"
__hyfi_user_config_path__: str = "config"
__hyfi_package_path__: str = Path(__file__).parent.parent.as_posix()
__hyfi_package_name__: str = os.path.basename(__hyfi_package_path__)
__hyfi_config_module__: str = f"{__hyfi_package_name__}.{__hyfi_config_dirname__}"
__hyfi_config_module_path__: str = f"pkg://{__hyfi_config_module__}"
__global_package_list__: Set[str] = {"cmd", "mode", "workflow"}

_batcher_instance_ = None


def __hyfi_version__() -> str:
    """
    Returns the version of HyFI.

    Returns:
        string containing the version of HyFI
    """
    from hyfi._version import __version__

    return __version__


class GlobalHyFIConfig(BaseModel):
    """Global configuration for HyFI

    Attributes:
    __package_name__ (str): The name of the package.
    __package_path__ (str): The path to the package root folder.
    __version__ (str): The version number of the package.
    __plugins__ (List[Any]): A list of plugins to load.
    __config_name__ (str): The name of the configuration module.
    __config_dirname__ (str): The name of the configuration directory.
    __user_config_path__ (str): The path to the user configuration directory.
    __dotenv_file__ (str): The name of the dotenv file.
    __secrets_dir__ (str): The name of the secrets directory.
    """

    __package_name__: str = __hyfi_package_name__
    __package_path__: str = __hyfi_package_path__
    __version__: str = __hyfi_version__()
    __plugins__: Optional[Dict[str, List[str]]] = None

    __config_name__: str = __hyfi_config_name__
    __config_dirname__: str = __hyfi_config_dirname__
    __user_config_path__: str = __hyfi_user_config_path__

    __dotenv_file__: str = ".env"
    __secrets_dir__: str = "./secrets"

    __verbosity__: Union[bool, int] = False

    _packages_: List[Tuple[str, str]] = [(__hyfi_package_path__, __hyfi_version__())]

    def initialize(
        self,
        package_path: str = __hyfi_name__,
        version: str = __hyfi_version__(),
        plugins: Optional[List[str]] = None,
        user_config_path: Optional[str] = None,
        config_dirname: Optional[str] = None,
        dotenv_file: Optional[str] = None,
        secrets_dir: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the global HyFI instance.

        This function should be called before any other HyFI function.

        A plugin is a python module which contains a configuration module.

        Be careful!
        It does not check if the plugin is importable.

        Args:
            package_path: Path to the package root folder. e.g. `./src/hyfi`
            version: Version of the package. e.g. `0.1.0`
            plugins: A list of plugins to load. e.g. `["hyfi.conf"]`
            user_config_path: Path to the user configuration directory. e.g. `./config`
            config_dirname: Name of the configuration directory. e.g. `conf`
            dotenv_file: Name of the dotenv file. e.g. `.env`
            secrets_dir: Name of the secrets directory. e.g. `secrets`
            **kwargs: Additional arguments to be set as attributes.
        """
        self.__package_path__ = package_path
        self.__version__ = version
        if package_path not in self._packages_:
            self._packages_.append((package_path, version))
        if plugins:
            self.__plugins__ = self.init_plugins(plugins)
        if user_config_path:
            self.__user_config_path__ = user_config_path
        if config_dirname:
            self.__config_dirname__ = config_dirname
        if dotenv_file:
            self.__dotenv_file__ = dotenv_file
        if secrets_dir:
            self.__secrets_dir__ = secrets_dir

        # for the future use
        for key, value in kwargs.items():
            key_name = (
                key if key.startswith("__") and key.endswith("__") else f"__{key}__"
            )
            if hasattr(self, key_name):
                setattr(self, key, value)
            else:
                logger.warning("Invalid key: %s", key)

    @property
    def dotenv_file(self) -> str:
        """Returns the name of the dotenv file."""
        if os.environ.get("DOTENV_FILE"):
            return os.environ.get("DOTENV_FILE")
        return self.__dotenv_file__

    @property
    def secrets_dir(self) -> str:
        """Returns the path of the secrets directory."""
        secret_dir = self.__secrets_dir__
        if not os.path.isdir(secret_dir):
            secret_dir = os.environ.get("HYFI_SECRETS_DIR", "")
        if os.path.isdir(secret_dir):
            secret_dir = (
                secret_dir
                if os.path.isabs(secret_dir)
                else os.path.join(os.getcwd(), secret_dir)
            )
        else:
            logger.debug("The secrets directory does not exist.")
            secret_dir = None
        return secret_dir

    @property
    def plugins(self) -> Optional[List[str]]:
        """Returns the list of plugins to load."""
        if self.__plugins__:
            caller_pkg_name = PKGs.get_next_level_caller_package_name()
            if caller_pkg_name in self.__plugins__:
                logger.debug(
                    "Loading plugins for %s: %s",
                    caller_pkg_name,
                    self.__plugins__[caller_pkg_name],
                )
                return self.__plugins__[caller_pkg_name]
        return None

    def init_plugins(self, plugins: List[str]) -> Dict[str, List[str]]:
        """Returns the list of plugins to load.
        A plugin is a python module which contains a configuration module.

        Args:
            plugins: List[str]: A list of plugins to load.

        Returns:
            Dict[str, List[str]]: A dictionary of plugins to load. ex) plugins = {'__package_name__': ['plugin1.conf', 'plugin2.conf']}
        """
        caller_pkg_name = PKGs.get_next_level_caller_package_name()
        _plugins = []
        for plugin in plugins:
            plugin = plugin.split(".")[0]
            if PKGs.is_importable(plugin):
                logger.debug("Plugin %s is importable.", plugin)
                config_module = f"{plugin}.{self.__config_dirname__}"
            else:
                logger.debug("Plugin %s is not importable.", plugin)
            _plugins.append(config_module)
        return {caller_pkg_name: _plugins}

    @property
    def package_name(self) -> str:
        """Returns the name of the package."""
        self.__package_name__ = os.path.basename(self.package_path)
        return self.__package_name__

    @property
    def package_path(self) -> str:
        """Returns the path to the package root folder.

        If there are multiple packages, the second package is returned.
        If there is only one package, the first package is returned. (default: hyfi)
        """
        if len(self._packages_) > 1:
            self.__package_path__ = self._packages_[1][0]
        else:
            self.__package_path__ = self._packages_[0][0]
        return self.__package_path__

    @property
    def version(self) -> str:
        """Returns the version number of the package."""
        if len(self._packages_) > 1:
            self.__version__ = self._packages_[1][1]
        else:
            self.__version__ = self._packages_[0][1]
        return self.__version__

    @property
    def config_dirname(self) -> str:
        """Returns the name of the configuration directory."""
        return self.__config_dirname__

    @property
    def config_root(self) -> str:
        """Returns the path to the configuration root directory."""
        return f"{self.package_path}/{self.config_dirname}"

    @property
    def config_module(self) -> str:
        """Returns the name of the configuration module."""
        return f"{self.package_name}.{self.config_dirname}"

    @property
    def config_module_path(self) -> str:
        """Returns the path to the configuration module."""
        return f"pkg://{self.config_module}"

    @property
    def config_name(self) -> str:
        """Returns the name of the configuration module."""
        return self.__config_name__

    @property
    def user_config_path(self) -> str:
        """Returns the path to the user configuration directory."""
        # if user_config_path is not an absolute path, make it absolute
        search_path = self.__user_config_path__
        if not os.path.isdir(search_path):
            search_path = os.environ.get("HYFI_USER_CONFIG_PATH", "")
        if os.path.isdir(search_path):
            self.__user_config_path__ = (
                search_path
                if os.path.isabs(search_path)
                else os.path.join(os.getcwd(), search_path)
            )
        else:
            logger.debug(
                "The user configuration directory does not exist: %s", search_path
            )
            self.__user_config_path__ = ""
        return self.__user_config_path__

    @property
    def hyfi_name(self) -> str:
        """Returns the name of HyFI package."""
        return __hyfi_name__

    @property
    def hyfi_config_module_path(self) -> str:
        """Returns the path to HyFI root folder"""
        return __hyfi_config_module_path__

    @property
    def hyfi_config_module(self) -> str:
        """Returns the name of HyFI default configuration module."""
        return __hyfi_config_module__

    @property
    def hyfi_config_name(self) -> str:
        """Returns the name of HyFI default configuration module."""
        return __hyfi_config_name__

    @property
    def hyfi_config_dirname(self) -> str:
        """Returns the name of HyFI default configuration directory."""
        return __hyfi_config_dirname__

    @property
    def hyfi_package_name(self) -> str:
        """Returns the name of HyFI package."""
        return __hyfi_package_name__

    @property
    def hyfi_package_path(self) -> str:
        """Returns the path to the package root folder."""
        return __hyfi_package_path__

    @property
    def hydra_version_base(self) -> str:
        """Returns the version of Hydra."""
        return __hydra_version_base__

    @property
    def hydra_default_config_group_value(self) -> str:
        """Returns the default config group value of Hydra."""
        return __hydra_default_config_group_value__

    @property
    def global_package_list(self) -> Set[str]:
        """Returns the list of global packages."""
        return __global_package_list__

    @property
    def verbosity(self) -> int:
        """Returns the verbosity level."""
        if os.environ.get("HYFI_VERBOSITY"):
            self.__verbosity__ = int(os.environ.get("HYFI_VERBOSITY", 0))
        return self.__verbosity__


__global_hyfi__ = GlobalHyFIConfig()


class GlobalHyFIResolver:
    @staticmethod
    def __hyfi_version__() -> str:
        """
        Returns the version of HyFI.

        Returns:
            string containing the version of HyFI
        """
        from hyfi._version import __version__

        return __version__

    def __hyfi_path__() -> str:
        """Returns the path to the HyFI root folder"""
        return __global_hyfi__.hyfi_package_path

    @staticmethod
    def __home_path__() -> str:
        """Returns the path to the user's home folder"""
        return Path.home().as_posix()

    @staticmethod
    def __app_version__() -> str:
        """
        Returns the version of App.

        Returns:
            string containing the version of App
        """

        return __global_hyfi__.version

    @staticmethod
    def __package_name__() -> str:
        """
        Returns the package name of the App

        Returns:
            string containing the package name of the App
        """

        return __global_hyfi__.package_name

    @staticmethod
    def __package_path__() -> str:
        """
        Returns the path to the App root folder

        Returns:
            string containing the path to the App root folder
        """

        return __global_hyfi__.package_path

    @staticmethod
    def __config_module_path__() -> str:
        """Global HyFI config path for the package to search for."""
        return __global_hyfi__.config_module_path

    @staticmethod
    def __user_config_path__() -> str:
        """Global HyFI user config path for the package to search for."""
        return __global_hyfi__.user_config_path



"""
This module contains the class for aggregating datasets.
"""
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class DSAggregate:
    def __init__(self):
        pass



"""Collage of images.""" ""
import math
import os
import textwrap
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from hyfi.composer import BaseModel, ConfigDict
from hyfi.utils.logging import LOGGING

from .utils import GUTILs

logger = LOGGING.getLogger(__name__)


class Collage(BaseModel):
    """Collage of images."""

    image: Image.Image
    width: int
    height: int
    ncols: int
    nrows: int
    filepath: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class COLLAGE:
    @staticmethod
    def collage(
        images_or_uris: Union[
            List[Union[str, Path, Image.Image]], str, Path, Image.Image
        ],
        output_file: Optional[Union[str, Path]] = None,
        num_cols: Optional[int] = 3,
        num_images: Optional[int] = 12,
        collage_width: int = 1200,
        padding: int = 10,
        background_color: str = "black",
        crop_to_min_size: bool = False,
        show_filename: bool = False,
        filename_offset: Tuple[int, int] = (5, 5),
        fontname: Optional[str] = None,
        fontsize: int = 12,
        fontcolor: str = "#000",
        **kwargs,
    ) -> Optional[Collage]:
        """
        Create a collage of images.
        """

        if not isinstance(images_or_uris, list):
            images_or_uris = [images_or_uris]
        images_or_uris = [
            uri if isinstance(uri, Image.Image) else str(uri) for uri in images_or_uris
        ]

        if num_images:
            num_images = min(num_images, len(images_or_uris))
        else:
            num_images = len(images_or_uris)
        if num_images < 1:
            raise ValueError("max_images must be greater than 0")
        num_images, num_cols, nrows = GUTILs.get_grid_size(
            num_images, num_cols=num_cols
        )
        logger.info(
            "Creating collage of %d images in %d columns and %d rows",
            num_images,
            num_cols,
            nrows,
        )
        img_width = collage_width // num_cols
        collage_width = num_cols * img_width + padding * (num_cols + 1)

        # load images
        images = GUTILs.load_images(
            images_or_uris[:num_images],
            resize_to_multiple_of=None,
            crop_to_min_size=crop_to_min_size,
            max_width=img_width,
            **kwargs,
        )
        filenames = [
            os.path.basename(image_or_uri) if isinstance(image_or_uri, str) else None
            for image_or_uri in images_or_uris[:num_images]
        ]
        # convert images
        images = [
            GUTILs.print_filename_on_image(
                image,
                print_filename=show_filename,
                filename=filename,
                filename_offset=filename_offset,
                fontname=fontname,
                fontsize=fontsize,
                fontcolor=fontcolor,
            )
            for image, filename in zip(images, filenames)
        ]

        collage = COLLAGE.get_grid_of_images(
            images, num_cols, padding, background_color=background_color
        )
        if output_file:
            output_file = str(output_file)
            os.makedirs(os.path.dirname(output_file), exist_ok=True)
            collage.image.save(output_file)
            collage.filepath = output_file
            logger.info("Saved collage to %s", output_file)
        return collage

    @staticmethod
    def label_collage(
        collage: Collage,
        collage_filepath=None,
        title: Optional[str] = None,
        title_fontsize: int = 10,
        xlabel=None,
        ylabel=None,
        xticklabels=None,
        yticklabels=None,
        xlabel_fontsize=12,
        ylabel_fontsize=12,
        dpi=100,
        fg_fontcolor="white",
        bg_color="black",
        caption=None,
        **kwargs,
    ) -> Collage:
        """
        Create a collage of images.
        """
        figsize = (collage.width / dpi, collage.height / dpi)
        ncols, nrows = collage.ncols, collage.nrows

        fig = plt.figure(figsize=figsize, dpi=dpi)
        fig.patch.set_facecolor(bg_color)  # type: ignore
        plt.imshow(np.array(collage.image))
        ax = plt.gca()
        plt.grid(False)
        if xlabel is None and ylabel is None:
            plt.axis("off")
        if title is not None:
            title = "\n".join(
                sum(
                    (
                        textwrap.wrap(
                            t, width=int(collage.width / 15 * 12 / title_fontsize)
                        )
                        for t in title.split("\n")
                    ),
                    [],
                )
            )
            ax.set_title(title, fontsize=title_fontsize, color=fg_fontcolor)
        if xlabel is not None:
            # plt.xlabel(xlabel, fontdict={"fontsize": xlabel_fontsize})
            ax.set_xlabel(xlabel, fontsize=xlabel_fontsize, color=fg_fontcolor)
        if ylabel is not None:
            # plt.ylabel(ylabel, fontdict={"fontsize": ylabel_fontsize})
            ax.set_ylabel(ylabel, fontsize=ylabel_fontsize, color=fg_fontcolor)
        if xticklabels is not None:
            # get ncols number of xticks from xlim
            xlim = ax.get_xlim()
            xticks = GUTILs.get_ticks_from_lim(xlim, ncols)
            ax.set_xticks(xticks, color=fg_fontcolor)
            xticklabels = [""] + xticklabels
            ax.set_xticklabels(
                xticklabels, fontsize=xlabel_fontsize, color=fg_fontcolor
            )
        if yticklabels is not None:
            # get nrows number of yticks from ylim
            ylim = ax.get_ylim()
            yticks = GUTILs.get_ticks_from_lim(ylim, nrows)
            ax.set_yticks(yticks, color=fg_fontcolor)
            yticklabels = [""] + yticklabels
            ax.set_yticklabels(
                yticklabels, fontsize=ylabel_fontsize, color=fg_fontcolor
            )

        plt.tight_layout()
        if caption is not None:
            plt.figtext(
                0.5,
                0.01,
                caption,
                wrap=True,
                horizontalalignment="center",
                fontsize=10,
                color=fg_fontcolor,
            )

        img = GUTILs.convert_figure_to_image(fig, dpi=dpi)
        img = GUTILs.scale_image(img, max_width=collage.width)
        plt.close()

        if collage_filepath is not None:
            collage_filepath = str(collage_filepath)
            os.makedirs(os.path.dirname(collage_filepath), exist_ok=True)
            # fig.savefig(collage_filepath, dpi=dpi, bbox_inches="tight", pad_inches=0)
            img.save(collage_filepath)
            # collage_image.save(collage_filepath)
            # log.info(f"Saved collage to {collage_filepath}")

        return Collage(
            image=img,
            filepath=collage_filepath,
            width=img.width,
            height=img.height,
            ncols=ncols,
            nrows=nrows,
        )

    @staticmethod
    def get_grid_of_images(
        images: List[Image.Image],
        num_cols: int = 3,
        padding: int = 10,
        background_color: str = "black",
    ) -> Collage:
        """
        Create a grid of images.
        """
        nrows = len(images) // num_cols
        assert len(images) == nrows * num_cols
        width, height = images[0].size
        grid_width = num_cols * width + padding * (num_cols + 1)
        grid_height = nrows * height + padding * (nrows + 1)
        collage = Image.new(
            "RGB", size=(grid_width, grid_height), color=background_color
        )
        for j, image in enumerate(images):
            x = j % num_cols
            y = j // num_cols
            collage.paste(
                image, (x * width + padding * (x + 1), y * height + padding * (y + 1))
            )
        return Collage(
            image=collage,
            width=grid_width,
            height=grid_height,
            ncols=num_cols,
            nrows=nrows,
        )

    @staticmethod
    def gallery_ndarray_images(
        array: np.ndarray,
        ncols: int = 7,
    ) -> np.ndarray:
        """
        Create a gallery of images from a numpy array.
        """
        nindex, height, width, intensity = array.shape
        nrows = nindex // ncols
        assert nindex == nrows * ncols
        return (
            array.reshape(nrows, ncols, height, width, intensity)
            .swapaxes(1, 2)
            .reshape(height * nrows, width * ncols, intensity)
        )

    @staticmethod
    def make_subplot_pages_from_images(
        images: List[Union[str, Path, np.ndarray, Image.Image]],
        num_images_per_page: int,
        num_cols: int,
        num_rows: Optional[int] = None,
        output_dir: Optional[Union[str, Path]] = None,
        output_file_format: str = "collage_p{page_num}.png",
        titles: Optional[List[str]] = None,
        title_fontsize: int = 10,
        title_color: str = "black",
        figsize: Optional[Tuple[float, float]] = None,
        width_multiple: float = 4,
        height_multiple: float = 2,
        sharex: bool = True,
        sharey: bool = True,
        squeeze: bool = True,
        dpi: int = 100,
        verbose: bool = False,
    ):
        """Make subplot pages from images.

        Args:
            images: List of images to be plotted.
            num_images_per_page: Number of images per page.
            num_cols: Number of columns.
            num_rows: Number of rows. If None, it will be calculated automatically.
            output_dir: Output directory.
            output_file_format: Output file format.
            titles: List of titles for each image.
            title_fontsize: Title fontsize.
            title_color: Title color.
            figsize: Figure size.
            width_multiple: Width multiple.
            height_multiple: Height multiple.
            sharex: Share x-axis.
            sharey: Share y-axis.
            squeeze: Squeeze.
            dpi: Dots per inch.
        """
        num_images = len(images)
        num_pages = math.ceil(num_images / num_images_per_page)
        for page_num in range(num_pages):
            start_idx = page_num * num_images_per_page
            end_idx = start_idx + num_images_per_page
            page_images = images[start_idx:end_idx]
            page_titles = titles[start_idx:end_idx] if titles else None
            page_output_file = (
                Path(output_dir) / output_file_format.format(page_num=page_num)
                if output_dir
                else None
            )
            if verbose:
                logger.info(
                    f"Making page {page_num + 1}/{num_pages} with {len(page_images)} images"
                )
                logger.info(f"Page titles: {page_titles}")
                logger.info(f"Page output file: {page_output_file}")
            COLLAGE.make_subplots_from_images(
                page_images,
                num_cols=num_cols,
                num_rows=num_rows,
                output_file=page_output_file,
                titles=page_titles,
                title_fontsize=title_fontsize,
                title_color=title_color,
                figsize=figsize,
                width_multiple=width_multiple,
                height_multiple=height_multiple,
                sharex=sharex,
                sharey=sharey,
                squeeze=squeeze,
                dpi=dpi,
                verbose=verbose,
            )

    @staticmethod
    def make_subplots_from_images(
        images: List[Union[str, Path, np.ndarray, Image.Image]],
        num_cols: int,
        num_rows: Optional[int] = None,
        output_file: Optional[Union[str, Path]] = None,
        titles: Optional[List[str]] = None,
        title_fontsize: int = 10,
        title_color: str = "black",
        figsize: Optional[Tuple[float, float]] = None,
        width_multiple: float = 4,
        height_multiple: float = 2,
        sharex: bool = True,
        sharey: bool = True,
        squeeze: bool = True,
        dpi: int = 100,
        verbose: bool = False,
    ):
        """Make subplots from images.

        Args:
            images: List of images to be plotted.
            num_cols: Number of columns.
            num_rows: Number of rows. If None, it will be calculated automatically.
            output_file: Output file path.
            titles: List of titles for each image.
            title_fontsize: Title fontsize.
            title_color: Title color.
            figsize: Figure size.
            sharex: Share x-axis or not.
            sharey: Share y-axis or not.
            squeeze: Squeeze or not.
            dpi: Dots per inch.
        """
        if num_rows is None:
            num_images = len(images)
            num_rows = math.ceil(num_images / num_cols)
        if figsize is None:
            figsize = (num_cols * width_multiple, num_rows * height_multiple)
        fig, axes = plt.subplots(
            num_rows,
            num_cols,
            figsize=figsize,
            sharex=sharex,
            sharey=sharey,
            squeeze=squeeze,
        )
        for i in range(num_cols * num_rows):
            ax = (
                axes[i % num_cols]
                if num_rows == 1
                else axes[i // num_cols, i % num_cols]
            )
            if i >= len(images):
                ax.set_visible(False)
                continue
            image = images[i]
            if isinstance(image, (str, Path)):
                image = GUTILs.load_image(image)
            ax.imshow(image)
            if titles:
                ax.set_title(titles[i], fontsize=title_fontsize, color=title_color)
            ax.axis("off")
        if output_file:
            GUTILs.save_adjusted_subplots(fig, output_file, dpi=dpi, verbose=verbose)



import io
from abc import abstractmethod
from typing import ClassVar, Optional, Tuple, Type, Union

import requests


class SchemeClient:
    """
    A client used for caching remote resources corresponding to URLs with a particular scheme.

    Subclasses must define the :attr:`scheme` class variable and implement
    :meth:`get_etag()` and :meth:`get_resource()`.

    .. important::
        Take care when implementing subclasses to raise the right error types
        from :meth:`get_etag()` and :meth:`get_resource()`.
    """

    recoverable_errors: ClassVar[Tuple[Type[BaseException], ...]] = (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
    )
    """
    Subclasses can override this to define error types that will be treated as recoverable.

    If ``cached_path()`` catches of one these errors while calling :meth:`get_etag()`, it
    will log a warning and return the latest cached version if there is one, otherwise
    it will propogate the error.
    """

    scheme: ClassVar[Union[str, Tuple[str, ...]]] = tuple()
    """
    The scheme or schemes that the client will be used for (e.g. "http").
    """

    def __init__(self, resource: str) -> None:
        self.resource = resource

    @abstractmethod
    def get_etag(self) -> Optional[str]:
        """
        Get the Etag or an equivalent version identifier associated with the resource.

        Returns
        -------
        ``Optional[str]``
            The ETag as a ``str`` or ``None`` if there is no ETag associated with
            the resource.

        Raises
        ------
        ``FileNotFoundError``
            If the resource doesn't exist.

        ``Recoverable error``
            Any error type defined in ``SchemeClient.recoverable_errors`` will
            be treated as a recoverable error.

            This means that when of these is caught by ``cached_path()``,
            it will look for cached versions of the given resource and return the
            latest version if there are any.

            Otherwise the error is propogated.

        ``Other errors``
            Any other error type can be raised. These errors will be treated non-recoverable
            and will be propogated immediately by ``cached_path()``.
        """
        raise NotImplementedError

    @abstractmethod
    def get_size(self) -> Optional[int]:
        """
        Get the size of the resource in bytes (if known).

        Returns
        -------
        ``Optional[int]``
            The size (in bytes).

        Raises
        ------
        ``FileNotFoundError``
            If the resource doesn't exist.

        ``Recoverable error``
            Any error type defined in ``SchemeClient.recoverable_errors`` will
            be treated as a recoverable error.

            This means that when of these is caught by ``cached_path()``, the size
            will be ignored.

        ``Other errors``
            Any other error type can be raised. These errors will be treated non-recoverable
            and will be propogated immediately by ``cached_path()``.
        """
        raise NotImplementedError

    @abstractmethod
    def get_resource(self, temp_file: io.BufferedWriter) -> None:
        """
        Download the resource to the given temporary file.

        Raises
        ------
        ``FileNotFoundError``
            If the resource doesn't exist.

        ``Other errors``
            Any other error type can be raised. These errors will be treated non-recoverable
            and will be propogated immediately by ``cached_path()``.
        """
        raise NotImplementedError



from .cached import CACHED
from .conf import CONFs
from .datasets import DATASETs
from .envs import ENVs
from .funcs import FUNCs
from .gpumon import GPUs
from .iolibs import IOLIBs
from .logging import LOGGING
from .notebooks import NBs
from .packages import PKGs
from .safeeval import SAFEEVAL, SafeEval


class UTILs(
    CACHED,
    CONFs,
    DATASETs,
    ENVs,
    FUNCs,
    GPUs,
    IOLIBs,
    LOGGING,
    NBs,
    PKGs,
    SAFEEVAL,
):
    def __init__(self) -> None:
        raise NotImplementedError("Use one of the static construction functions")


__all__ = [
    "CONFs",
    "DATASETs",
    "ENVs",
    "FUNCs",
    "GPUs",
    "IOLIBs",
    "LOGGING",
    "NBs",
    "PKGs",
    "UTILs",
    "SAFEEVAL",
    "SafeEval",
]



from enum import Enum
from os import PathLike
from pathlib import PosixPath, WindowsPath
from typing import Dict, List, Mapping, Sequence, Union

from omegaconf import DictConfig, ListConfig

IntSeq = Sequence[int]
ListLike = (ListConfig, List, Sequence)
DictLike = (DictConfig, Dict, Mapping)
DictKeyType = Union[str, bytes, int, Enum, float, bool]
PathOrStr = Union[str, PathLike]
PathLikeType = Union[str, PosixPath, WindowsPath]



"""
Filter datasets. Slice, sample, and filter datasets.
"""
import random
from typing import List, Optional, Sequence, Union

import numpy as np
import pandas as pd
from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict
from datasets.splits import Split

from hyfi.utils.logging import LOGGING

from .save import DSSave
from .types import DatasetLikeType

logger = LOGGING.getLogger(__name__)


class DSSlice:
    @staticmethod
    def sample_dataset(
        data: DatasetLikeType,
        split: Optional[Union[str, Split]] = None,
        sample_size: Union[int, float] = 100,
        sample_seed: int = 42,
        randomize: bool = True,
        num_heads: Optional[int] = 1,
        num_tails: Optional[int] = 1,
        verbose: bool = False,
    ) -> Dataset:
        """Sample a dataset."""
        if not isinstance(data, Dataset or DatasetDict):
            if split is None:
                raise ValueError(
                    "Please provide a split name when sampling a DatasetDict."
                )
            data = data[split]
        if sample_seed > 0:
            random.seed(sample_seed)
        if sample_size < 1:
            sample_size = int(len(data) * sample_size)
        if randomize:
            idx = random.sample(range(len(data)), sample_size)
        else:
            idx = range(sample_size)

        data = data.select(idx)
        logger.info("Sampling done.")
        if verbose:
            if num_heads:
                num_heads = min(num_heads, len(data))
                print(data[:num_heads])
            if num_tails:
                num_tails = min(num_tails, len(data))
                print(data[-num_tails:])

        return data

    @staticmethod
    def filter_and_sample_data(
        data: pd.DataFrame,
        queries: Optional[Union[str, List[str]]] = None,
        sample_size: Optional[Union[int, float]] = None,
        sample_seed: int = 42,
        output_dir: str = ".",
        sample_filename: Optional[str] = None,
        train_filename: Optional[str] = "train.parquet",
        discard_filename: Optional[str] = None,
        returning_data: str = "train",
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Filter the data to remove short documents and create a sample for analysis

        Args:
            data (pd.DataFrame): The data to filter and sample
            queries (Optional[Union[str, List[str]]], optional): The queries to filter by. Defaults to None.
            sample_size (Optional[Union[int, float]], optional): The sample size. Defaults to None.
            sample_seed (int, optional): The seed for sampling. Defaults to 42.
            output_dir (str, optional): The output directory. Defaults to ".".
            sample_filename (Optional[str], optional): The sample filename. Defaults to None.
            train_filename (Optional[str], optional): The train filename. Defaults to "train.parquet".
            discard_filename (Optional[str], optional): The discard filename. Defaults to "discard.parquet".
            returning_data (str, optional): The data to return. Defaults to "train" ('train', 'sample', 'discard', or 'original').
            verbose (bool, optional): Whether to print the sample. Defaults to False.

        Returns:
            pd.DataFrame: The original data
        """
        # Filter by queries
        if queries:
            data_ = DSSlice.filter_data_by_queries(
                data.copy(), queries, verbose=verbose
            )
        else:
            logger.warning("No query specified")
            data_ = data.copy()

        # Create a sample for analysis
        sample = None
        if sample_size:
            sample = DSSlice.sample_data(
                data_,
                sample_size_per_group=sample_size,
                sample_seed=sample_seed,
                verbose=verbose,
            )
            if sample_filename:
                DSSave.save_dataframes(
                    sample,
                    data_file=f"{output_dir}/{sample_filename}",
                    verbose=verbose,
                )
            if verbose:
                print(sample.head())

        # Create a train set
        train = data_[~data_.index.isin(sample.index)] if sample is not None else data_
        if train_filename:
            DSSave.save_dataframes(
                train,
                data_file=f"{output_dir}/{train_filename}",
                verbose=verbose,
            )
        if verbose:
            print(train.head())

        # Create a discard set
        if sample is not None:
            discard = data[
                ~data.index.isin(train.index) & ~data.index.isin(sample.index)
            ]
        else:
            discard = data[~data.index.isin(train.index)]
        if discard_filename:
            DSSave.save_dataframes(
                discard,
                data_file=f"{output_dir}/{discard_filename}",
                verbose=verbose,
            )
        if verbose:
            print(discard.head())
        logger.info(
            "Created %d samples, %d train samples, and %d discard samples",
            sample.shape[0] if sample is not None else 0,
            train.shape[0],
            discard.shape[0],
        )

        if returning_data == "train":
            return train
        elif returning_data == "sample" and sample is not None:
            return sample
        elif returning_data == "discard":
            return discard
        return data

    @staticmethod
    def filter_data_by_queries(
        data: pd.DataFrame,
        queries: Union[str, List[str]],
        verbose=False,
    ) -> pd.DataFrame:
        """
        Filter the data by queries

        Args:
            data: The data to filter
            queries: The queries to filter the data by
            verbose: Whether to print verbose logs

        Returns:
            The filtered data

        Examples:
            >>> import pandas as pd
            >>> data = pd.DataFrame({"text": ["hello world", "hello", "world"]})
            >>> filter_by_queries(data, "text.str.split().str.len() >= 2")
                text
            0  hello world
            >>> filter_by_queries(data, ["text.str.contains('hello')", "text.str.contains('world')"])
                text
            0  hello world
            2        world
        """
        if isinstance(queries, str):
            queries = [queries]

        for qry in queries:
            logger.info("filtering data by %s", qry)
            n_docs = data.shape[0]
            data = data.query(qry, engine="python")
            if verbose:
                logger.info("filtered %d documents", n_docs - data.shape[0])
        return data

    @staticmethod
    def sample_data(
        data: pd.DataFrame,
        sample_size_per_group: Union[float, int],
        sample_seed: int = 123,
        group_by: Optional[str] = None,
        value_col: Optional[str] = None,
        remove_columns: Optional[List[str]] = None,
        verbose: bool = False,
    ) -> pd.DataFrame:
        """
        Sample data from a dataframe

        Args:
            data: The dataframe to sample from
            sample_size_per_group: The number or fraction of samples to take from each group
            sample_seed: Random seed for sampling
            group_by: Column to group the data by, default is None (no grouping)
            value_col: Column containing the value to sample by, default is None (only for grouped sampling to count the number of samples per group)
            remove_columns: Columns to remove from sampled dataframe, default is None (keep all columns)
            verbose: Print verbose logs

        Returns:
            The sampled dataframe

        Examples:
            >>> import pandas as pd
            >>> data = pd.DataFrame({"text": ["hello world", "hello", "world"]})
            >>> sample_data(data, 1)
                text
            0  hello world
        """
        if verbose:
            logger.info("Sampling data")

        if not group_by:
            _sample = (
                data.sample(frac=sample_size_per_group, random_state=sample_seed)
                if sample_size_per_group < 1
                else data.sample(n=sample_size_per_group, random_state=sample_seed)
            )
        elif sample_size_per_group < 1:
            _sample = data.groupby(group_by, group_keys=False).apply(
                lambda x: x.sample(frac=sample_size_per_group, random_state=sample_seed)
            )
        else:
            _sample = data.groupby(group_by, group_keys=False).apply(
                lambda x: x.sample(
                    n=min(len(x), sample_size_per_group), random_state=sample_seed
                )
            )
        _sample.reset_index(drop=True, inplace=True)
        if remove_columns:
            logger.info("Removing columns: %s", remove_columns)
            _sample = _sample.drop(columns=remove_columns)
        if verbose:
            logger.info("Total rows in data: %s", len(data))
            logger.info("Total rows in sample: %s", len(_sample))
            print(_sample.head())

            if group_by and value_col:
                grp_all = data.groupby(group_by)[value_col].count().rename("population")
                grp_sample = (
                    _sample.groupby(group_by)[value_col].count().rename("sample")
                )
                grp_dists = pd.concat([grp_all, grp_sample], axis=1)
                print(grp_dists)

        return _sample

    @staticmethod
    def split_dataframe(
        data,
        indices_or_sections: Union[int, Sequence[int]],
        verbose: bool = False,
    ) -> List[pd.DataFrame]:
        """
        Split a dataframe into multiple dataframes

        Args:
            data (pd.DataFrame): dataframe to split
            indices_or_sections (int or sequence of ints): if int, number of chunks to split the dataframe into

        Returns:
            List[pd.DataFrame]: list of dataframes

        Examples:
            >>> import pandas as pd
            >>> data = pd.DataFrame({"text": ["hello world", "hello", "world"]})
            >>> split_dataframe(data, 2)
            [        text
            0  hello world,
            1        hello,        text
            2  world]

        """

        if verbose:
            logger.info("Splitting dataframe into %s", indices_or_sections)
        return np.array_split(data, indices_or_sections)
